# (c) Copyright Riverlane 2020-2025.
"""Abstract class to define base functionality of API implementation."""

from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import numpy.typing as npt
import stim

from deltakit_explorer.enums._api_enums import APIEndpoints
from deltakit_explorer.types import (
    Decoder,
    DecodingResult,
    DetectionEvents,
    LeakageFlags,
    Measurements,
    NoiseModel,
    ObservableFlips,
    QubitCoordinateToDetectorMapping,
)
from deltakit_explorer.types._experiment_types import QECExperimentDefinition


class APIClient(ABC):  # pragma: nocover
    """Abstract class which presents and interface
    for an API"""

    def __init__(self, base_url: str, request_timeout: int):
        super().__init__()
        self.request_timeout = request_timeout
        if base_url and base_url[-1] == "/":
            base_url = base_url[:-1]
        self.base_url = base_url

    @abstractmethod
    def set_token(self, token: str, validate: bool = True):
        """Persist the token, set it to env var and a file.
        Validate it.

        Args:
            token (str): auth token.
            validate (bool): you may switch off validation explicitly.
        """
        raise NotImplementedError()

    @abstractmethod
    def execute(
        self,
        query_name: APIEndpoints,
        variable_values: dict,
        request_id: str,
    ) -> dict[str, Any]:
        """
        Executes a query by its name and
        returns server response as a python dict.

        Args:
            query_name (APIEndpoints): name of the query to execute.
            variable_values (dict): arguments for the query.
            request_id (str): identifier of the request.

        Returns:
            dict[str, Any]: Server response.
        """
        raise NotImplementedError()

    @abstractmethod
    def kill(self, request_id: str) -> int:
        """Kill a decoding task submitted to a client.
        Decoding methods report a request ID. Calling this
        method with an ID tries to kill all CPU
        workers associated with the task. Method returns the number
        of CPU workers it could terminate. If the task is already finished,
        the method should return 0.

        Args:
            request_id (str):
                ID of the requests which started the decoding task.

        Returns:
            int: number of workers killed
        """
        raise NotImplementedError()

    @abstractmethod
    def generate_circuit(
        self,
        experiment_definition: QECExperimentDefinition,
        request_id: str
    ) -> str:
        """Generate a circuit based on the experiment definition.

        Args:
            experiment_definition (QECExperimentDefinition):
                Definition of the quantum error correction experiment.
            request_id (str): Identifier of the request.

        Returns:
            str: Generated circuit in stim format."""
        raise NotImplementedError()

    @abstractmethod
    def simulate_circuit(
        self,
        stim_circuit: str | stim.Circuit,
        shots: int,
        request_id: str
    ) -> tuple[Measurements, LeakageFlags | None]:
        """Simulate a circuit and return measurements and leakage flags.

        Args:
            stim_circuit (str | stim.Circuit): Circuit to simulate.
            shots (int): Number of shots for the simulation.
            request_id (str): Identifier of the request.

        Returns:
            tuple[Measurements, LeakageFlags | None]: Tuple containing measurements and optional leakage flags."""
        raise NotImplementedError()

    @abstractmethod
    def add_noise(self, stim_circuit: str | stim.Circuit, noise_model: NoiseModel, request_id: str) -> str:
        """Add noise to a circuit based on the provided noise model.

        Args:
            stim_circuit (str | stim.Circuit): Circuit to which noise will be added.
            noise_model (NoiseModel): Noise model to apply.
            request_id (str): Identifier of the request.

        Returns:
            str: Noisy circuit in stim format. May contain leakage (not stim-compatible), if SI1000NoiseModel is used."""
        raise NotImplementedError()

    @abstractmethod
    def decode(
        self,
        detectors: DetectionEvents,
        observables: ObservableFlips,
        decoder: Decoder,
        noisy_stim_circuit: str | stim.Circuit,
        leakage_flags: LeakageFlags | None,
        request_id: str,
    ) -> DecodingResult:
        """Decode a noisy circuit with given detectors and observables.

        Args:
            detectors (DetectionEvents): DetectionEvents to use for decoding.
            observables (ObservableFlips): ObservableFlips to use for decoding.
            decoder (Decoder): Decoder to use.
            noisy_stim_circuit (str | stim.Circuit): Noisy circuit to decode.
            leakage_flags (LeakageFlags | None): Optional leakage flags.
            request_id (str): Identifier of the request.

        Returns:
            DecodingResult:
                Result of the decoding, which includes predictions,
                fails count, and CPU time spent on decoding.

        """
        raise NotImplementedError()

    @abstractmethod
    def defect_rates(
        self,
        detectors: DetectionEvents,
        stim_circuit: str | stim.Circuit,
        request_id: str,
    ) -> dict[tuple[float, ...], list[float]]:
        """Get defect rates for detectors.

        Args:
            detectors (DetectionEvents):
                DetectionEvents values from simulations to compute defect rates.
            stim_circuit (str | stim.Circuit): Circuit to analyse.
            request_id (str): Identifier of the request."""
        raise NotImplementedError()

    @abstractmethod
    def get_correlation_matrix_for_trimmed_data(
        self,
        detectors: DetectionEvents,
        noise_floor_circuit: str | stim.Circuit,
        use_default_noise_model_edges: bool,
        request_id: str,
    ) -> tuple[npt.NDArray[np.float64], QubitCoordinateToDetectorMapping]:
        """Get correlation matrix for detectors.

        Args:
            detectors (DetectionEvents): Detector data to analyse.
            noise_floor_circuit (str | stim.Circuit): Circuit with the minimal noise.
            use_default_noise_model_edges (bool): Whether to use default noise model edges.
            request_id (str): Identifier of the request.

        Returns:
            tuple[npt.NDArray[np.float64], QubitCoordinateToDetectorMapping]:
                Tuple containing the correlation matrix and a mapping of qubit coordinates to detectors.
        """
        raise NotImplementedError()

    @abstractmethod
    def trim_circuit_and_detectors(
        self,
        stim_circuit: str | stim.Circuit,
        detectors: DetectionEvents,
        request_id: str
    ) -> tuple[str, DetectionEvents]:
        """Trim a circuit and detectors to remove qubits and detectors irrelevant to decoding problem.

        Args:
            stim_circuit (str | stim.Circuit): Circuit to trim.
            detectors (DetectionEvents): DetectionEvents to trim.
            request_id (str): Identifier of the request.

        Returns:
            tuple[str, DetectionEvents]: Tuple containing the trimmed circuit in stim format and the trimmed detectors
        """
        raise NotImplementedError()



# (c) Copyright Riverlane 2020-2025.
"""
This module includes functions merging multiple circuits into a
single circuit to be executed on a single QPU in parallel.
"""


import numpy as np
from deltakit_circuit import (
    Circuit,
    Detector,
    GateLayer,
    Observable,
    Qubit,
    ShiftCoordinates,
)

# According to the purpose and business logic, content of this file
# may violate pylint rules on cyclomatic complexity. Still, code
# complexity requirement should be enforced for the new code.
# pylint: disable=too-many-branches,too-many-nested-blocks,too-many-statements
# pylint: disable=too-many-return-statements,too-many-boolean-expressions


def parallelise_disjoint_circuits(circuits: list[Circuit]) -> Circuit:
    """
    Parallelise a list of circuits, each of which must act on distinct qubits.
    We assume it is preferable to perform gates of the same type in the same
    layer and thus this is done if possible, but not if it would involve
    making the total number of gate layers in the parallelised circuit larger
    than the maximum number of gate layers in any of the input circuits.
    None of the circuits may contain NoiseLayers. Only one circuit may
    contain annotation layers and only one circuit may contain nested Circuit
    layers - these need not be the same circuit. If one circuit contains
    annotations, the other circuits may not contain measurement gates. The
    circuits must also have the same number of iterations.

    Example
    -------
    Input:
        Circuit 1:
            Layer 1: CNOT(0, 1)
            Layer 2: H(1), H(2)
        Circuit 2:
            Layer 1: H(3)

    Output:
        Circuit:
            Layer 1: CNOT(0, 1)
            Layer 2: H(1), H(2), H(3)


    Parameters
    ----------
    circuits : List[Circuit]
        List of circuits to be parallelised.

    Returns
    -------
    Circuit
        Parallelised circuit.
    """
    if len(circuits) == 0:
        return Circuit()

    # Check for unequal numbers of iterations
    iterations = circuits[0].iterations
    if any(circuit.iterations != iterations for circuit in circuits):
        msg = "Circuits to be parallelised must have the same number of iterations."
        raise ValueError(msg)

    def _any_annotations(circuit: Circuit) -> bool:
        for layer in circuit.layers:
            if isinstance(layer, (Detector, Observable, ShiftCoordinates)):
                return True
            if isinstance(layer, Circuit) and _any_annotations(layer):
                return True
        return False

    def _any_circuit_layers(circuit: Circuit) -> bool:
        return any(isinstance(layer, Circuit) for layer in circuit.layers)

    def _any_measurements(circuit: Circuit) -> bool:
        if len(circuit.measurement_gates) > 0:
            return True
        return any(
            _any_measurements(layer)
            for layer in circuit.layers
            if isinstance(layer, Circuit)
        )

    # Check if any circuit contains NoiseLayers
    if any(circuit.is_noisy for circuit in circuits):
        msg = "Circuits to be parallelised may not contain NoiseLayers."
        raise ValueError(msg)

    # Check which circuits contain annotations
    any_annotations = [_any_annotations(circuit) for circuit in circuits]

    if sum(any_annotations) == 1:
        annotations_circ = any_annotations.index(True)
        if any(
            _any_measurements(circuit)
            for circuit in circuits
            if circuit != circuits[annotations_circ]
        ):
            msg = (
                "If one circuit to be parallelised contains annotations, "
                "no other circuit can contain measurements."
            )
            raise ValueError(msg)
    elif sum(any_annotations) > 1:
        msg = "Only one circuit to be parallelised can contain annotations."
        raise ValueError(msg)

    # Check how many circuits contain a nested circuit
    if sum(_any_circuit_layers(circuit) for circuit in circuits) > 1:
        msg = "Only one circuit to be parallelised can contain a nested Circuit."
        raise ValueError(msg)

    # In future, this can be simplified to the length of non-recursive gate_layers
    circuit_lengths = [
        len([layer for layer in circuit.layers if isinstance(layer, GateLayer)])
        for circuit in circuits
    ]
    total_length = max(circuit_lengths)

    # Get layers for longest circuit
    max_circ_ind = circuit_lengths.index(total_length)
    max_circ = circuits[max_circ_ind]

    # Initialise lists to store the layers and separately the gate types in each layer
    parallelised_layers = []
    parallelised_layers_gate_types = []

    # Add gate types from longest circuit to gate type list and non-GateLayers to
    # layer list
    for layer in max_circ.layers:
        if isinstance(layer, GateLayer):
            parallelised_layers.append(GateLayer())
            parallelised_layers_gate_types.append(
                {type(gate) for gate in layer.gates}
            )
        else:
            parallelised_layers.append(layer)
            parallelised_layers_gate_types.append(set())

    # Initialise set to record qubits
    qubits: set[Qubit] = set()

    # Loop over circuits
    for icirc, (circuit, new_length) in enumerate(zip(circuits, circuit_lengths)):
        # update qubits
        new_qubits = circuit.qubits
        if len(qubits.union(new_qubits)) != len(qubits) + len(new_qubits):
            msg = "Circuits to be parallelised do not act on distinct qubits."
            raise ValueError(msg)
        qubits = qubits.union(new_qubits)

        # set variable to record previous layer in parallelised circuit into which
        # gates were added or where an entire layer was inserted
        par_previous_layer_ind = -1

        # set variable to record the gate layer index of the layer to be inserted
        new_gate_layer_ind = -1

        # set variable to record the gate layer index in the parallelised circuit
        par_gate_layer_ind = -1

        for new_layer in circuit.layers:
            if not isinstance(new_layer, GateLayer):
                par_previous_layer_ind += 1
                if icirc != max_circ_ind:
                    parallelised_layers.insert(par_previous_layer_ind, new_layer)
                    parallelised_layers_gate_types.insert(par_previous_layer_ind, set())
            else:
                new_gate_layer_ind += 1
                for par_layer_ind in range(
                    par_previous_layer_ind + 1, len(parallelised_layers)
                ):
                    par_layer = parallelised_layers[par_layer_ind]
                    par_layer_gate_types = parallelised_layers_gate_types[par_layer_ind]
                    if not isinstance(par_layer, GateLayer):
                        continue

                    # check if number of gate layers left is equal to number available or if
                    # all gates are already in layer
                    par_gate_layer_ind += 1
                    if (
                        new_length - new_gate_layer_ind
                        == total_length - par_gate_layer_ind
                    ) or {
                        type(gate) for gate in new_layer.gates
                    } <= par_layer_gate_types:
                        # add gates to layer
                        par_layer.add_gates(new_layer.gates)

                        # record gate types
                        if icirc != max_circ_ind:
                            par_layer_gate_types.update(
                                {type(gate) for gate in new_layer.gates}
                            )
                        par_previous_layer_ind += par_layer_ind - par_previous_layer_ind
                        break

    return Circuit(parallelised_layers)


def parallelise_same_length_circuits(circuits: list[Circuit]) -> Circuit:
    """
    Parallelise a list of circuits, each of which must be the same length and
    act on distinct qubits in a particular layer. The circuits must also consist
    of GateLayers only.

    Parameters
    ----------
    circuits : List[Circuit]
        List of circuits to be parallelised.

    Returns
    -------
    Circuit
        Parallelised circuit.
    """
    if len(circuits) == 0:
        return Circuit()

    # Check for unequal numbers of iterations
    iterations = circuits[0].iterations
    if any(circuit.iterations != iterations for circuit in circuits):
        msg = "Circuits to be parallelised must have the same number of iterations."
        raise ValueError(msg)

    if any(
        not all(isinstance(layer, GateLayer) for layer in circuit.layers)
        for circuit in circuits
    ):
        msg = "Circuits can only be parallelised if they contain only GateLayers."
        raise ValueError(msg)

    circuit_lengths = [len(circuit.layers) for circuit in circuits]
    num_layers = circuit_lengths[0]
    if not np.all(np.array(circuit_lengths) == num_layers):
        msg = "Circuits must all be the same length."
        raise ValueError(msg)

    # Begin parallelised circuit construction with first circuit
    parallelised_circuit = Circuit(
        [GateLayer(layer.gates) for layer in circuits[0].layers]
    )

    # Loop over remaining circuits
    for circuit in circuits[1:]:
        for ilayer, layer in enumerate(circuit.layers):
            parallelised_circuit.layers[ilayer].add_gates(layer.gates)

    return parallelised_circuit



# (c) Copyright Riverlane 2020-2025.
"""
This module stores an implementation of the rotated planar code.
"""
# pylint: disable=too-many-branches, too-many-boolean-expressions
import itertools

from deltakit_circuit import PauliX, PauliZ, Qubit
from deltakit_circuit._basic_types import Coord2D, Coord2DDelta

from deltakit_explorer.codes._planar_code._planar_code import PlanarCode, ScheduleType
from deltakit_explorer.codes._schedules._rotated_planar_code_schedules import (
    RotatedPlanarCodeSchedules,
)
from deltakit_explorer.codes._schedules._schedule_order import (
    ScheduleOrder,
    get_x_and_z_schedules,
)


class RotatedPlanarCode(PlanarCode):
    r"""
    Class representing the default rotated planar code. For the default rotated
    planar code, Z logical is vertical (i.e. Z boundaries are horizontal) and X
    logical is horizontal (i.e. X boundaries are vertical). The default patch
    has a weight 2 horizontal Z-stabiliser on the left top corner. See the graph
    below showing a default 3x3 rotated code.

    This class also contains methods that help set up circuits for memory
    experiments.

    By default, the rotated planar code will have the following layout:

    .. code-block:: text

       6├              /   Z   \
        │            /           \
       5├           ○ ----------- ○ ----------- ○ \
        │           │             │             │   \
       4├           │      X      │      Z      │     X
        │           │             │             │   /
       3├         / ○ ----------- ○ ----------- ○ /
        │       /   │             │             │
       2├    X      │      Z      │      X      │
        │       \   │             │             │
       1├         \ ○ ----------- ○ ----------- ○
        │                          \           /
       0├                            \   Z   /
        │
        └----┴------┴------┴------┴------┴------┴------┴
             0      1      2      3      4      5      6

    Parameters
    ----------
    width : int
        Has to be >= 2. The patch is built on a rectangular grid of points
        of size width x height.
    height : int
        Has to be >= 2. The patch is built on a rectangular grid of points
        of size width x height.
    schedule_type : ScheduleType, optional
        The scheduling type to measure stabilisers.
        By default, ScheduleType.SIMULTANEOUS.
    schedule_order : ScheduleOrder, optional
        Enum specifying the order of the schedule, from a number of available
        options. By default, ScheduleOrder.HORIZONTALLY_REFLECTED, which,
        in case `top_bumps_are_z == True`, will perform an N-shape schedule on
        the X stabilisers and a Z-shape schedule on the Z stabilisers, both
        starting in the top left and finishing in the bottom right. In the other
        case `top_bumps_are_z == False`, it will do N-shape on Z and Z-shape on
        X stabilisers.
    use_ancilla_qubits : bool
        Whether or not to use ancilla qubits for the stabilisers. By default, True.
    shift : Coord2DDelta, optional
        A vector by which to shift all of the qubit coordinates, by adding it to each
        qubit coordinate.
        By default, the 0 vector: Coord2DDelta(0, 0).
    horizontal_bump_with_top_left: bool, optional.
        A bump is a weight-2 stabiliser.
        True if and only if the patch is stabilised by a weight-2
        plaquette supported on top_left_vertex and
        top_left_vertex + Coord2DDelta(2, 0). By default, True.
    top_bumps_are_z: bool, optional.
        Specifies whether the weight-2 stabilisers on the top side of
        the patch are of Z type or not. This fixes the direction of the logical operators;
        if True then logical Z is vertical and logical X is horizontal.
        By default, True.
    """

    def __init__(
        self,
        width: int,
        height: int,
        schedule_type: ScheduleType = ScheduleType.SIMULTANEOUS,
        schedule_order: ScheduleOrder = ScheduleOrder.HORIZONTALLY_REFLECTED,
        use_ancilla_qubits: bool = True,
        shift: Coord2DDelta = Coord2DDelta(0, 0),
        horizontal_bump_with_top_left: bool = True,
        top_bumps_are_z: bool = True,
    ):
        self._horizontal_bump_with_top_left = horizontal_bump_with_top_left
        self._top_bumps_are_z = top_bumps_are_z
        self._x_type_has_N_shape = (  # pylint: disable=invalid-name
            top_bumps_are_z
        )
        x_schedule, z_schedule = get_x_and_z_schedules(
            RotatedPlanarCodeSchedules,
            schedule_order,
            x_type_has_N_shape=self._x_type_has_N_shape,
        )

        self._perform_css_checks = False

        if self._top_bumps_are_z:
            self.x_distance = width
            self.z_distance = height
        else:
            self.x_distance = height
            self.z_distance = width

        super().__init__(
            width=width,
            height=height,
            untransformed_x_schedule=x_schedule,
            untransformed_z_schedule=z_schedule,
            schedule_type=schedule_type,
            use_ancilla_qubits=use_ancilla_qubits,
            shift=shift,
        )

    def _calculate_default_patch_data_qubits(self) -> set[Qubit]:
        """Calculates the untransformed data qubits for a default rotated
        code patch.

        Returns
        -------
        Set[Qubit]
            Data qubits for a standard rotated code patch.
        """
        return {
            Qubit(Coord2D(x, y))
            for (x, y) in itertools.product(
                range(1, 2 * self.width, 2),
                range(1, 2 * self.height, 2),
            )
        }

    def _calculate_default_patch_ancilla_qubits(
        self,
    ) -> tuple[set[Qubit], set[Qubit]]:
        """Calculates the untransformed ancilla qubits for a default rotated
        code patch.

        Returns
        -------
        Tuple[Set[Qubit], Set[Qubit]]
            All ancilla qubits of opposite type from the top weight-2
            stabilizers for a standard rotated code patch. All ancilla qubits
            of same type as the top weight-2 stabilizers for a standard rotated
            code patch.
        """
        # Collect all qubits where an X-ancilla could be placed in the default
        # `top_bumps_are_z == True` case.
        possible_x_ancilla_qubits = {
            Qubit(Coord2D(x, y))
            for x, y in itertools.product(
                range(0, 2 * self.width + 1, 2),
                range(2, 2 * self.height - 1, 2),
            )
        }
        default_x_ancilla_qubits = {
            anc
            for anc in possible_x_ancilla_qubits
            if (anc.unique_identifier.x + anc.unique_identifier.y - 2 * self.height) % 4
            == 0
        }

        # Collect all qubits where a Z-ancilla could be placed in the default
        # `top_bumps_are_z == True` case.
        possible_z_ancilla_qubits = {
            Qubit(Coord2D(x, y))
            for x, y in itertools.product(
                range(2, 2 * self.width - 1, 2),
                range(0, 2 * self.height + 1, 2),
            )
        }
        default_z_ancilla_qubits = {
            anc
            for anc in possible_z_ancilla_qubits
            if (anc.unique_identifier.x + anc.unique_identifier.y - 2 * self.height) % 4
            == 2
        }

        if self._horizontal_bump_with_top_left:
            x_ancilla_qubits = default_x_ancilla_qubits
            z_ancilla_qubits = default_z_ancilla_qubits
        else:
            x_ancilla_qubits = possible_x_ancilla_qubits.difference(
                default_x_ancilla_qubits
            )
            z_ancilla_qubits = possible_z_ancilla_qubits.difference(
                default_z_ancilla_qubits
            )

        return x_ancilla_qubits, z_ancilla_qubits

    def _calculate_untransformed_all_qubits(
        self,
    ) -> tuple[set[Qubit], set[Qubit], set[Qubit]]:
        data_qubits = self._calculate_default_patch_data_qubits()
        (
            x_ancilla_qubits,
            z_ancilla_qubits,
        ) = self._calculate_default_patch_ancilla_qubits()

        return (
            (x_ancilla_qubits, z_ancilla_qubits, data_qubits)
            if self._top_bumps_are_z
            else (z_ancilla_qubits, x_ancilla_qubits, data_qubits)
        )

    def _calculate_untransformed_logical_operators(
        self,
    ) -> tuple[tuple[set[PauliX], ...], tuple[set[PauliZ], ...]]:
        horiz_op_pauli, vert_op_pauli = (
            (PauliX, PauliZ) if self._top_bumps_are_z else (PauliZ, PauliX)
        )
        vert_op = {
            vert_op_pauli(Qubit(Coord2D(1, y))) for y in range(1, 2 * self.height, 2)
        }
        horiz_op = {
            horiz_op_pauli(Qubit(Coord2D(x, 2 * self.height - 1)))
            for x in range(1, 2 * self.width, 2)
        }

        return (
            ((horiz_op,), (vert_op,))
            if self._top_bumps_are_z
            else ((vert_op,), (horiz_op,))
        )



# (c) Copyright Riverlane 2020-2025.
"""
This module contains common implementation parts for stabiliser codes.
CSSCode class derives from StabiliserCode.
"""

import itertools
from abc import ABC, abstractmethod
from collections.abc import Collection, Iterable, Sequence

from deltakit_circuit import Qubit
from deltakit_circuit._basic_types import PauliBasis
from deltakit_circuit._qubit_identifiers import PauliGate

from deltakit_explorer.codes._css._css_stage import CSSStage
from deltakit_explorer.codes._css._stabiliser_helper_functions import (
    pauli_gates_to_stim_pauli_string,
)
from deltakit_explorer.codes._stabiliser import Stabiliser


class StabiliserCode(ABC):
    """
    Class representing an abstract stabiliser code.

    This class also contains methods that help set up circuits for memory
    experiments.
    """

    def __init__(
        self,
        stabilisers: Sequence[Iterable[Stabiliser]],
        x_logical_operators: Sequence[Collection[PauliGate]] | None = None,
        z_logical_operators: Sequence[Collection[PauliGate]] | None = None,
        use_ancilla_qubits: bool = True,
        check_logical_operators_are_independent: bool = False,
    ):
        self._check_logical_operators_are_independent = (
            check_logical_operators_are_independent
        )
        self._stabilisers = tuple(
            tuple(simultaneous_stabilisers) for simultaneous_stabilisers in stabilisers
        )
        self._use_ancilla_qubits = use_ancilla_qubits
        if not hasattr(self, "_data_qubits"):
            self._data_qubits = self._calculate_data_qubits()
        self._ancilla_qubits = self._calculate_ancilla_qubits()

        if (x_logical_operators is None) != (z_logical_operators is None):
            msg = "Either both or neither of the logical operators should be provided."
            raise ValueError(msg)

        self._x_logical_operators = (
            tuple(frozenset(x_logical) for x_logical in x_logical_operators)
            if x_logical_operators is not None
            else None
        )
        self._z_logical_operators = (
            tuple(frozenset(z_logical) for z_logical in z_logical_operators)
            if z_logical_operators is not None
            else None
        )

    @abstractmethod
    def encode_logical_zeroes(self) -> CSSStage:
        r"""
        Set up the CSSStage which represents the encoding of a logical
        :math:`\ket{0}\dots\ket{0}`.

        Returns
        -------
        CSSStage
            CSSStage representing the encoding of a logical
            :math:`\ket{0}\dots\ket{0}`.
        """

    @abstractmethod
    def encode_logical_pluses(self) -> CSSStage:
        r"""
        Set up the CSSStage which represents the encoding of a logical
        :math:`\ket{+}\dots\ket{+}`

        Returns
        -------
        CSSStage
            CSSStage representing the encoding of a logical
            :math:`\ket{+}\dots\ket{+}`.
        """

    def _calculate_data_qubits(self) -> set[Qubit]:
        """
        Calculate all data qubits and return a set of them.

        Returns
        -------
        Set[Qubit]
            Set of all data qubits.
        """
        data_qubits: set[Qubit] = set()
        for simultaneous_stabilisers in self._stabilisers:
            for stab in simultaneous_stabilisers:
                data_qubits.update(stab.data_qubits)
        return data_qubits

    def _calculate_ancilla_qubits(self) -> set[Qubit]:
        """
        Calculate ancilla qubits.

        Returns
        -------
        Set[Qubit]
            Set of all ancilla qubits.
        """
        if not self._use_ancilla_qubits:
            return set()

        ancilla_attrs = set(
            itertools.chain.from_iterable(
                {stab.ancilla_qubit for stab in simultaneous_stabilisers}
                for simultaneous_stabilisers in self._stabilisers
            )
        )
        if None in ancilla_attrs:
            msg = (
                "In order to perform syndrome extraction using ancilla qubits, all "
                "the Stabilisers must have ancilla qubits defined."
            )
            raise ValueError(msg)
        return ancilla_attrs

    def measure_stabilisers(self, num_rounds: int) -> CSSStage:
        """
        Set up the CSSStage which represents measuring the stabilisers for num_rounds
        rounds.

        Parameters
        ----------
        num_rounds : int
            Number of rounds to measure the stabilisers for.

        Returns
        -------
        CSSStage
            CSSStage representing measuring the stabiliser for num_rounds
            rounds.
        """
        return CSSStage(
            num_rounds=num_rounds,
            stabilisers=self._stabilisers,
            use_ancilla_qubits=self._use_ancilla_qubits,
        )

    @abstractmethod
    def measure_z_logicals(self) -> CSSStage:
        """
        Set up the CSSStage which represents measuring all logical Z operators
        simultaneously.

        Returns
        -------
        CSSStage
             CSSStage representing measuring the logical Z.
        """

    @abstractmethod
    def measure_x_logicals(self) -> CSSStage:
        """
        Set up the CSSStage which represents measuring all logical X operators
        simultaneously.

        Returns
        -------
        CSSStage
             CSSStage representing measuring the logical X.
        """

    @property
    def stabilisers(self) -> tuple[tuple[Stabiliser, ...], ...]:
        """
        Code stabilisers.

        Returns
        -------
        Tuple[Tuple[Stabiliser, ...], ...]
            Code stabilisers.
        """
        return self._stabilisers

    @property
    def qubits(self) -> set[Qubit]:
        """
        All code qubits, both ancilla and data.

        Returns
        -------
        Set[Qubit]
            A set of qubits.
        """
        return self._ancilla_qubits.union(self._data_qubits)

    @property
    def data_qubits(self) -> set[Qubit]:
        """
        All code data qubits.

        Returns
        -------
        Set[Qubit]
            A set of data qubits.
        """
        return self._data_qubits

    @property
    def ancilla_qubits(self) -> set[Qubit]:
        """
        All code ancilla qubits.

        Returns
        -------
        Set[Qubit]
            A set of ancilla qubits.
        """
        return self._ancilla_qubits

    @property
    def use_ancilla_qubits(self) -> bool:
        """
        Whether ancilla qubits are used for syndrome extraction.

        Returns
        -------
        bool
            True, if ancillas are used.
        """
        return self._use_ancilla_qubits

    @property
    def x_logical_operators(self) -> tuple[frozenset[PauliGate], ...]:
        """
        All X logical operators for the code.

        Returns
        -------
        Tuple[FrozenSet[PauliGate]]
            A tuple of frozensets of `PauliGate`s, with each frozenset
            corresponding to one logical X operator.
        """
        return (
            self._x_logical_operators
            if self._x_logical_operators is not None
            else ()
        )

    @x_logical_operators.setter
    def x_logical_operators(self, new_x_logicals: Sequence[Collection[PauliGate]]):
        """
        Set the value of x_logical_operators with new Pauli-products.
        This setter will validate that each new logical:
        (a) Commutes with all the other new logicals,
        (b) Commutes with all stabilisers,
        (c) Anti-commute with exactly one Z-logical (at the same index), and commute with
        the others In the event that any check fails, an error is raised. It is also
        checked whether there are as many new logicals as existing logicals, and a
        ValueError is thrown if there is a difference.
        If the code's logicals are being set for the first time, and the code currently
        has no logicals, the above criteria are not checked, but will be checked for any
        subsequent setting of logicals.

        Parameters
        ----------
        new_x_logicals : Sequence[Collection[PauliGate]]
            New X-logicals, to be validated and set if valid.
        """
        self._validate_new_logicals(
            new_x_logicals,
            PauliBasis.Z,
        )
        self._x_logical_operators = tuple(frozenset(log) for log in new_x_logicals)

    @property
    def z_logical_operators(self) -> tuple[frozenset[PauliGate], ...]:
        """
        All Z logical operators for the code.

        Returns
        -------
        Tuple[FrozenSet[PauliGate]]
            A tuple of frozensets of `PauliGate`s, with each frozenset
            corresponding to one logical Z operator.
        """
        return (
            self._z_logical_operators
            if self._z_logical_operators is not None
            else ()
        )

    @z_logical_operators.setter
    def z_logical_operators(self, new_z_logicals: Sequence[Collection[PauliGate]]):
        """
        Set the value of z_logical_operators with new Pauli-products.
        This setter will validate that each new logical:
        (a) Commutes with all the other new logicals,
        (b) Commutes with all stabilisers,
        (c) Anti-commutes with exactly one Z-logical (at the same index), and commutes with
        the others.
        In the event that any check fails, an error is raised. It is also
        checked whether there are as many new logicals as existing logicals, and a
        ValueError is thrown if there is a difference.
        If the code's logicals are being set for the first time, and the code currently
        has no logicals, the above criteria are not checked, but will be checked for any
        subsequent setting of logicals.

        Parameters
        ----------
        new_z_logicals : Sequence[Collection[PauliGate]]
            New Z-logicals, to be validated and set if valid.
        """
        self._validate_new_logicals(
            new_z_logicals,
            PauliBasis.X,
        )
        self._z_logical_operators = tuple(frozenset(log) for log in new_z_logicals)

    def overwrite_logicals(
        self,
        new_x_logicals: Sequence[Collection[PauliGate]],
        new_z_logicals: Sequence[Collection[PauliGate]],
    ):
        """
        Method to allow for updating both X and Z logicals simultaneously. For instance,
        the user may want to define a new logical that does not anti-commute with the
        current other type logical, but that will anti-commute with the new logical to
        be specified. Trying to do this in sequence will raise an error, so this method
        allows both to be updated simultaneously.
        Validation is performed on all the new logical operators.

        Parameters
        ----------
        new_x_logicals : Sequence[Collection[PauliGate]]
            New X-logicals, to be validated and set if valid.
        new_z_logicals : Sequence[Collection[PauliGate]]
            New Z-logicals, to be validated and set if valid.
        """
        # Use private attribute to bypass validation for first setting of X
        self._x_logical_operators = tuple(frozenset(log) for log in new_x_logicals)
        # Set Z using the setter, to validate the new Z against the new X
        self.z_logical_operators = tuple(frozenset(log) for log in new_z_logicals)
        # Then also set X again, where it will now be compared to the Z that has been
        # set, as X could possibly have been invalid (e.g, not commute with stabilisers)
        self.x_logical_operators = self._x_logical_operators

    def _validate_new_logicals(
        self,
        new_logicals: Sequence[Collection[PauliGate]],
        opposite_logical_type: PauliBasis,
    ):
        """
        Given a sequence of new logicals, a code, stabilisers of the
        code and the opposite type logicals to the ones being validated, check whether the
        provided new logicals
        (a) Commutes with all the other new logicals,
        (b) Commutes with all stabilisers,
        (c) Anti-commutes with exactly one Z-logical (at the same index), and commutes with
        the others.
        In the event that any check fails, an error is raised. It is also
        checked whether there are as many new logicals as existing logicals, and a
        ValueError is thrown if there is a difference.
        If the code's logicals are being set for the first time, and the code currently
        has no logicals, the above criteria are not checked, but will be checked for any
        subsequent setting of logicals.

        Parameters
        ----------
        new_logicals : Sequence[Collection[PauliGate]]
            The new logical operators to validate.
        opposite_logical_type : PauliBasis
            The basis of the logical operators against which the new logicals will be
            checked for anti-commutativity.
        """
        stabilisers = self.stabilisers
        data_qubit_to_index = {qubit: i for i, qubit in enumerate(self.data_qubits)}
        opposite_type_logicals = (
            self.z_logical_operators
            if opposite_logical_type == PauliBasis.Z
            else self.x_logical_operators
        )
        # Allow there being 0 opposite logicals, in the case where a code is initialised without
        # logicals, then they are added later.
        if len(opposite_type_logicals) > 0 and len(new_logicals) != len(
            opposite_type_logicals
        ):
            msg = (
                "There must be as many new logicals as existing logicals,"
                f" but there are {len(new_logicals)}"
                f" while there should be {len(opposite_type_logicals)}"
            )
            raise ValueError(msg)
        if any(len(log) == 0 for log in new_logicals):
            msg = "Logicals cannot be weight 0"
            raise ValueError(msg)

        # Convert stabilisers and logicals to stim.PauliString for commutation checks
        stabilisers_as_pauli_strings = [
            pauli_gates_to_stim_pauli_string(stab.paulis, data_qubit_to_index)
            for stab in itertools.chain.from_iterable(stabilisers)
        ]
        new_logs_as_pauli_string = [
            pauli_gates_to_stim_pauli_string(logical, data_qubit_to_index)
            for logical in new_logicals
        ]
        opposite_logs_as_pauli_string = [
            pauli_gates_to_stim_pauli_string(logical, data_qubit_to_index)
            for logical in opposite_type_logicals
        ]

        for i, new_log_as_pauli_string in enumerate(new_logs_as_pauli_string):
            # Check logical commutes with other logicals of same type
            if not all(
                new_log_as_pauli_string.commutes(other_new_logical)
                for other_new_logical in new_logs_as_pauli_string[i + 1 :]
            ):
                msg = f"New logical at index {i} anti-commutes with other new logicals"
                raise ValueError(msg)

            # Check logical commutes with all stabilisers
            if not all(
                new_log_as_pauli_string.commutes(stab)
                for stab in stabilisers_as_pauli_strings
            ):
                msg = f"New logical at index {i} anti-commutes with stabilisers"
                raise ValueError(msg)

            # Check for anti-commutation if this isn't the first time setting logicals
            if (
                len(opposite_type_logicals) > 0
                and self._check_logical_operators_are_independent
            ):
                # Check new logical at index i anti-commutes with opposite-type logical
                # at index i
                if new_log_as_pauli_string.commutes(opposite_logs_as_pauli_string[i]):
                    msg = (
                        f"New logical at index {i} must anti-commute with opposite type"
                        f" logical at index {i}"
                    )
                    raise ValueError(msg)

                # Check new logical at index i commutes with all the other opposite-type
                # logical operators
                if not all(
                    new_log_as_pauli_string.commutes(log)
                    for log in opposite_logs_as_pauli_string[:i]
                    + opposite_logs_as_pauli_string[(i + 1) :]
                ):
                    msg = (
                        f"New logical at index {i} anti-commutes with opposite-type"
                        f" logicals other than at index {i}"
                    )
                    raise ValueError(msg)



# (c) Copyright Riverlane 2020-2025.
"""
Classes and functions to encapsulate syndrome extraction schedules for Planar codes.
This module is not currently public; this __init__.py file is a historical artefact
and can be removed, adjusting imports within other `deltakit_explorer` modules
accordingly.
"""
from deltakit_explorer.codes._schedules._rotated_planar_code_schedules import (
    RotatedPlanarCodeSchedules,
)
from deltakit_explorer.codes._schedules._schedule_order import (
    ScheduleClass,
    ScheduleOrder,
    get_x_and_z_schedules,
)
from deltakit_explorer.codes._schedules._unrotated_planar_code_schedules import (
    UnrotatedPlanarCodeSchedules,
)

__all__ = [s for s in dir() if not s.startswith("_")]



from collections.abc import Callable, Mapping, Sequence

import numpy as np
import numpy.typing as npt
from deltakit_circuit._circuit import Circuit
from deltakit_decode.analysis import RunAllAnalysisEngine

from deltakit_explorer.analysis.error_budget._generation import (
    generate_decoder_managers_for_lambda,
)
from deltakit_explorer.analysis.error_budget._memory import (
    MemoryGenerator,
    PreComputedMemoryGenerator,
    get_rotated_surface_code_memory_circuit,
)
from deltakit_explorer.analysis.error_budget._parameters import SamplingParameters
from deltakit_explorer.analysis.error_budget._post_processing import (
    compute_lambda_and_stddev_from_results,
)


def inverse_lambda_at(
    noise_model: Callable[[Circuit, npt.NDArray[np.floating]], Circuit],
    noise_parameters: npt.NDArray[np.floating] | Sequence[float],
    num_rounds_by_distances: Mapping[int, Sequence[int]],
    sampling_parameters: SamplingParameters = SamplingParameters(),
    memory_generator: MemoryGenerator
    | Mapping[int, Mapping[int, Circuit]] = get_rotated_surface_code_memory_circuit,
) -> tuple[float, float]:
    """Compute 1 / Λ.

    Warning:
        This is a helper function to compute 1 / Λ when you need a **single**
        evaluation.
        For error budgeting, :func:`~deltakit_explorer.analysis.error_budget.get_error_budget`
        will be able to parallelise more efficiently, while also performing several
        checks and optimisations.

    Args:
        noise_model (Callable[[Circuit, npt.NDArray[np.floating]], Circuit]): a callable
            adding noise to the provided circuit, according to the parameters provided.
        noise_parameters (npt.NDArray[numpy.floating] | Sequence[float]): valid
            parameters to forward to ``noise_model`` representing the point at which the
            gradient should be computed.
        num_rounds_by_distances (Mapping[int, Sequence[int]]): a mapping from each code
            distance that should be tested to the number of rounds that should be
            sampled in order to estimate the logical error-probability per round, to
            ultimately get 1 / Λ.
        sampling_parameters: additional parameters relating to the sampling tasks used to
            estimate 1 / Λ indirectly.
        memory_generator (MemoryGenerator): a callable that can generate a memory
            experiment. The resulting circuit will go through the provided
            ``noise_model`` for different values of the noise parameters.

    Returns:
        the estimation of 1 / Λ along with the standard deviation of the estimation as
        a 2-tuple.
    """
    if isinstance(memory_generator, Mapping):
        memory_generator = PreComputedMemoryGenerator(memory_generator)

    point = np.asarray(noise_parameters).reshape((-1, 1))

    # Create unique identifiers for noise parameters that will be used to discriminate between them
    # in the CSV file storing the simulation results.
    noise_parameter_names = [str(i) for i in range(point.size)]

    decoder_managers = generate_decoder_managers_for_lambda(
        point,
        noise_model,
        num_rounds_by_distances,
        sampling_parameters.max_workers,
        memory_generator=memory_generator,
        noise_parameter_names=noise_parameter_names,
    )
    engine = RunAllAnalysisEngine(
        experiment_name="Estimating 1 / Λ",
        decoder_managers=decoder_managers,
        max_shots=sampling_parameters.max_shots,
        batch_size=sampling_parameters.batch_size,
        # Early stopping when we have a low-enough standard deviation
        loop_condition=RunAllAnalysisEngine.loop_until_observable_rse_below_threshold(
            sampling_parameters.lep_target_rse,
            sampling_parameters.lep_computation_min_fails,
        ),
        num_parallel_processes=sampling_parameters.max_workers,
    )
    report = engine.run()
    lambdas, lambda_stddevs = compute_lambda_and_stddev_from_results(
        point, noise_parameter_names, num_rounds_by_distances, report
    )
    lambda_reciprocals = 1 / lambdas
    lambda_reciprocal_stddevs = np.abs(lambda_stddevs / lambdas**2)

    return float(lambda_reciprocals[0, 0]), float(lambda_reciprocal_stddevs[0, 0])



# (c) Copyright Riverlane 2020-2025.
"""
This module defines Enums used in hardware-aware noise modelling.
These enums are used in `qpus/_qpu_factory.py` and
`noise_models/_noise_factory.py` modules.
"""

from enum import Enum


class DrawingColours(Enum):
    """
    Enumeration of colours used in drawing patch diagrams.

    The options are:
    - X_COLOUR
    - Z_COLOUR
    - DATA_QUBIT_COLOUR
    - ANCILLA_QUBIT_COLOUR
    """

    X_COLOUR = "#d3d1c0"
    Z_COLOUR = "#3ccbda"
    DATA_QUBIT_COLOUR = "#006f62"
    ANCILLA_QUBIT_COLOUR = "#ff7500"



# (c) Copyright Riverlane 2020-2025.
"""
This module contains standalone functions related to observables used to generate
stabiliser code circuits.
"""
from collections.abc import Iterable, Mapping, Sequence

from deltakit_circuit import MeasurementRecord, Observable, Qubit
from deltakit_circuit.gates import ONE_QUBIT_MEASUREMENT_GATES, _MeasurementGate


def _construct_observables(
    observable_definitions: Mapping[int, Iterable[_MeasurementGate | Qubit]],
    measurements: Sequence[_MeasurementGate],
) -> list[Observable]:
    """
    Constructs Observable objects from a set of observable definitions and
    a list of lookback measurements. For each observable definition,
    finds the lookback index of the latest measurement matching each element in the
    definition.

    Parameters
    ----------
    observable_definitions : Mapping[int, Iterable[_MeasurementGate | Qubit]]
        Dictionary describing the observable definitions from which to construct
        the Observable objects. Its keys are observable indices and its values
        specify the set of measurements that should make up the corresponding
        observable to construct. These can either be (1) a qubit, in which case any
        measurement involving that qubit will get matched, or (2) a measurement,
        in which case only a measurement in that specific basis will get matched.
    measurements : Sequence[_MeasurementGate]
        The list of all measurements that have previously occurred ordered in the same
        way than they occur in the circuit. These must go far enough in the past such
        that all elements in the observable definitions are covered. Will raise a
        ValueError if no match is found.

    Returns
    -------
    List[Observable]
        Constructed list of observables.
    """
    observables: list[Observable] = []
    num_measurements = len(measurements)
    measurement_qubits = [
        measurement.qubit
        if isinstance(measurement, tuple(ONE_QUBIT_MEASUREMENT_GATES))
        else None
        for measurement in measurements
    ]
    for (
        observable_ind,
        observable_measurements,
    ) in observable_definitions.items():
        measurement_records = []
        for observable_measurement in observable_measurements:
            meas_ind = -num_measurements
            try:
                if isinstance(observable_measurement, Qubit):
                    meas_ind += measurement_qubits.index(observable_measurement)
                else:
                    meas_ind += measurements.index(observable_measurement)
            except ValueError as ve:
                msg = (
                    f"{observable_measurement} has not been measured and "
                    "thus its measurement result cannot be included in a "
                    "logical observable."
                )
                raise ValueError(msg) from ve
            measurement_records.append(MeasurementRecord(meas_ind))

        observables.append(Observable(observable_ind, measurement_records))

    return observables



# (c) Copyright Riverlane 2020-2025.
"""Module contains functions to perform analysis of data assets to perform efficient
data routing between client and service."""
from __future__ import annotations

import re

from stim import Circuit

from deltakit_explorer.enums._api_enums import DataFormat
from deltakit_explorer.types._data_string import DataString


def get_binary_data_size(width: int, shots: int, data_format: DataFormat) -> int:
    """Get number of bytes to represent a binary table (width x shots)
    using the data_format.

    Args:
       width (int): bits per shot.
       shots (int): number of shots.
       data_format (DataFormat): how the data is packed.

    Returns:
        int: size in memory, in bytes
    """
    if data_format == DataFormat.B8:
        # each shot is a bit
        length = (width + 7) // 8
    elif data_format == DataFormat.F01:
        # + endline
        length = width + 1
    elif data_format == DataFormat.CSV:
        # commas + endline
        length = 2 * width
    else:
        msg = f"Format {data_format} is not supported for binary data"
        raise NotImplementedError(
            msg
        )
    return length * shots


def get_decoding_response_size_estimate(
    circuit: str | Circuit,
    shots: int,
    data_format: DataFormat,
) -> int:
    """Get an estimation for a decoding response, in bytes.
    Size may slightly vary based on moving parts
    of the request. Estimation is an upper bound.

    Args:
       circuit (Union[str, Circuit]): a circuit, for which detectors are decoded.
       shots (int): number of shots to decode.
       data_format (DataFormat): how the data is packed.

    Returns:
        int: size of response content, in bytes
    """
    if isinstance(circuit, Circuit):
        width = circuit.num_observables
    else:
        width = len(re.findall("OBSERVABLE_INCLUDE", circuit))
    bytes_data_size = get_binary_data_size(width, shots, data_format)
    duck_size = len(DataString.empty) + 2 * bytes_data_size
    # all other field, including shots, errors, times, quotes
    # empirical upper bound estimate
    wrapper_size = 200
    return wrapper_size + duck_size


def get_decoding_request_size_estimate(
    circuit: str | Circuit,
    shots: int,
    data_format: DataFormat,
) -> int:
    """Get an estimation for a decoding request, in bytes.
    Size may slightly vary based on moving parts
    of the request. Estimation is an upper bound.

    Args:
       circuit (Union[str, Circuit]): a circuit, for which detectors are decoded.
       shots (int): number of shots to decode.
       data_format (DataFormat): how the data is packed.

    Returns:
        int: size of request content, in bytes
    """
    if isinstance(circuit, Circuit):
        width_det = circuit.num_detectors
        width_obs = circuit.num_observables
        width_leakage = 0
    else:
        width_det = len(re.findall(r"DETECTOR\(", circuit))
        width_obs = len(re.findall(r"OBSERVABLE_INCLUDE", circuit))
        width_leakage = 0
        for line in circuit.splitlines(False):
            line = line.strip()
            # add all measurements and leakages
            if line.startswith("HERALD_LEAKAGE_EVENT"):
                # there is 1 operator followed by arguments
                width_leakage += len(line.split()) - 1
    duck_size = 2 * (
        len(DataString.empty)  # used twice
        + get_binary_data_size(width_det, shots, data_format)
        + get_binary_data_size(width_obs, shots, data_format)
        + get_binary_data_size(width_leakage, shots, data_format)
    )
    # all other field, including variable names, formats, ...
    # empirical upper bound estimate
    wrapper_size = 200
    return wrapper_size + duck_size


def get_simulation_response_size_estimate(
    circuit: str | Circuit,
    shots: int,
    data_format: DataFormat,
) -> int:
    """Get an estimation for a simulation response, in bytes.
    Size may slightly vary based on moving parts
    of the response. Estimation is an upper bound.

    Args:
       circuit (Union[str, Circuit]): a circuit to simulate.
       shots (int): number of shots to simulate.
       data_format (DataFormat): how the data is packed.

    Returns:
        int: size of response content, in bytes
    """
    if isinstance(circuit, Circuit):
        width_mmt = circuit.num_measurements
        width_leak = 0
    else:
        width_mmt, width_leak = 0, 0
        for line in circuit.splitlines(False):
            line = line.strip()
            # add all measurements and leakages
            if line.startswith("M"):
                width_mmt += len(line.split()) - 1
            elif line.startswith("HERALD_LEAKAGE_EVENT"):
                # there is 1 operator followed by arguments
                width_leak += len(line.split()) - 1
    duck_size = 2 * (
        len(DataString.empty)
        + get_binary_data_size(width_mmt, shots, data_format)
        + get_binary_data_size(width_leak, shots, data_format)
    )
    # all other field, including formats, quotes, ...
    # empirical upper bound estimate
    wrapper_size = 200
    return wrapper_size + duck_size


def has_leakage(stim_circuit: str) -> bool:
    """Check is circuit has any leakage-related instructions.
    If it has, if should only be stored in strings and processed
    as client.

    Args:
        stim_circuit (str): string representation of STIM circuit.

    Returns:
        bool: True if any leakage-related instruction are there.
    """
    leakage_evidence = {"LEAKAGE", "HERALD_LEAKAGE_EVENT", "RELAX"}
    stim_text = str(stim_circuit)
    stim_lines = stim_text.splitlines()
    return any(
        line.strip().startswith(prefix)
        for line in stim_lines
        for prefix in leakage_evidence
    )



# (c) Copyright Riverlane 2020-2025.
"""
This module consists of individual circuit optimisation functions.
"""


import itertools

from deltakit_circuit import (
    Circuit,
    Detector,
    GateLayer,
    MeasurementRecord,
    NoiseLayer,
    Observable,
    ShiftCoordinates,
)
from deltakit_circuit.gates import I

# According to the purpose and business logic, content of this file
# may violate pylint rules on cyclomatic complexity. Still, code
# complexity requirement should be enforced for the new code.
# pylint: disable=too-many-branches,too-many-nested-blocks,too-many-statements
# pylint: disable=too-many-return-statements,too-many-boolean-expressions


def merge_layers(circuit: Circuit, break_repeat_blocks: bool = False) -> Circuit:
    """
    Merge adjacent gate layers if they act on distinct qubits. Appropriate reindexing
    of measurement records also occurs. Once initial merging has occurred, merging of
    a layer from outside a repeat block with a layer from inside a repeat block will
    be considered. A maximum of one repeat of the circuit may be removed from
    the end of the repeat block, if the break_repeat_blocks option is specified and
    if the next layer after the repeat block is a gate layer. This function requires
    the circuit to have no noise layers.

    Parameters
    ----------
    circuit : Circuit
        Input circuit with layers to be merged.
    break_repeat_blocks : bool, optional
        If True, a maximum of one repeat from the end of a repeat block will be
        "broken out" of the block in order to facilitate merging with a layer
        outside the repeat block. By default, False.

    Returns
    -------
    Circuit
        Circuit with merged layers.
    """

    def _calculate_offset_annotation_layers(
        annotation_layers_with_offsets: list[
            tuple[Detector | Observable | ShiftCoordinates, int]
        ],
    ) -> list[Detector | Observable | ShiftCoordinates]:
        """
        Construct a list of annotations with shifted measurement indices.

        Parameters
        ----------
        annotation_layers_with_offsets : List[
            Union[Detector, Observable, ShiftCoordinates], int]
        ]
            Input annotation layers with their respective offsets to be applied to
            the measurement record measurement indices.

        Returns
        -------
        List[Union[Detector, Observable, ShiftCoordinates]]
            Annotation layers with measurement record indices shifted.
        """

        layers = []
        for annotation, offset in annotation_layers_with_offsets:
            if isinstance(annotation, ShiftCoordinates):
                layers.append(annotation)
            else:
                measurement_records = {
                    MeasurementRecord(measurement_record.lookback_index - offset)
                    for measurement_record in annotation.measurements
                }
                if isinstance(annotation, Detector):
                    new_annotation = Detector(
                        measurements=measurement_records,
                        coordinate=annotation.coordinate,
                    )
                else:
                    new_annotation = Observable(
                        measurements=measurement_records,
                        observable_index=annotation.observable_index,
                    )
                layers.append(new_annotation)

        return layers

    iterations = circuit.iterations

    merged_layers = []
    current_gate_layer = GateLayer()
    current_annotations_and_offsets: list[
        tuple[Detector | Observable | ShiftCoordinates, int]
    ] = []
    for layer in circuit.layers:
        if isinstance(layer, GateLayer):
            # try to merge layer with current_gate_layer
            if current_gate_layer.qubits.isdisjoint(layer.qubits):
                current_gate_layer.add_gates(layer.gates)
                current_annotations_and_offsets = [
                    (annotation, measurement_offset + len(layer.measurement_gates))
                    for (
                        annotation,
                        measurement_offset,
                    ) in current_annotations_and_offsets
                ]
            else:
                merged_layers.append(current_gate_layer)
                merged_layers += _calculate_offset_annotation_layers(
                    current_annotations_and_offsets
                )
                current_gate_layer = GateLayer(layer.gates)
                current_annotations_and_offsets = []
        elif isinstance(layer, Circuit):
            # apply current_gate_layer and annotations before considering repeat block
            if len(current_gate_layer.gates) > 0:
                merged_layers.append(current_gate_layer)
                current_gate_layer = GateLayer()
            merged_layers += _calculate_offset_annotation_layers(
                current_annotations_and_offsets
            )
            current_annotations_and_offsets = []
            merged_layers.append(merge_layers(layer, break_repeat_blocks))
        elif isinstance(layer, NoiseLayer):
            msg = "Layer merge cannot be carried out on a circuit with noise layers."
            raise ValueError(msg)
        else:
            # add annotation to current annotation layers
            current_annotations_and_offsets.append((layer, 0))

    # add remaining layers to end of circuit
    if len(current_gate_layer.gates) != 0:
        merged_layers.append(current_gate_layer)
    merged_layers += _calculate_offset_annotation_layers(
        current_annotations_and_offsets
    )

    if not break_repeat_blocks:
        return Circuit(merged_layers, iterations=iterations)

    # now remove one repeat from the repeat block if it is beneficial to do so
    new_merged_layers = []
    skip_layer = False
    for layer, next_layer in itertools.pairwise(merged_layers):
        if skip_layer:
            skip_layer = False
            continue

        if isinstance(layer, Circuit) and isinstance(next_layer, GateLayer):
            nested_circuit_layers = layer.layers
            gate_layer = GateLayer(next_layer.gates)
            # try to merge nested circuit and following gate layer
            merged_nested_circuit_and_layer = merge_layers(
                Circuit([*nested_circuit_layers, gate_layer]), break_repeat_blocks
            )
            if len(merged_nested_circuit_and_layer.layers) <= len(
                nested_circuit_layers
            ):
                # if merge is successful, remove one repeat from the nested circuit
                new_merged_layers.append(
                    Circuit(nested_circuit_layers, iterations=layer.iterations - 1)
                )
                new_merged_layers += merged_nested_circuit_and_layer.layers
                skip_layer = True
            else:
                new_merged_layers.append(layer)
        else:
            new_merged_layers.append(layer)

    if not skip_layer:
        new_merged_layers.append(merged_layers[-1])

    return Circuit(new_merged_layers, iterations=iterations)


def remove_identities(circuit: Circuit) -> Circuit:
    """
    Remove identity gates from a circuit.

    Parameters
    ----------
    circuit : Circuit
        Circuit from which to remove identity gates.

    Returns
    -------
    Circuit
        Circuit with identity gates removed.
    """
    new_layers = []
    for layer in circuit.layers:
        if isinstance(layer, Circuit):
            new_layer = remove_identities(layer)
            if len(new_layer.layers) > 0:
                new_layers.append(new_layer)
        elif isinstance(layer, GateLayer):
            new_layer = GateLayer(
                [gate for gate in layer.gates if not isinstance(gate, I)]
            )
            if len(new_layer.gates) > 0:
                new_layers.append(new_layer)
        else:
            new_layers.append(layer)

    return Circuit(new_layers, iterations=circuit.iterations)



# (c) Copyright Riverlane 2020-2025.
"""This file contains implementation of an API for accessing
the Deltakit service.
"""

from __future__ import annotations

import os

import numpy as np
import numpy.typing as npt
import stim

from deltakit_explorer import simulation
from deltakit_explorer._api._api_client import APIClient
from deltakit_explorer._api._api_v2_client import APIv2Client
from deltakit_explorer._api._auth import TOKEN_VARIABLE
from deltakit_explorer._api._gql_client import GQLClient
from deltakit_explorer._utils import _utils as utils
from deltakit_explorer._utils._decorators import (
    validate_and_split_decoding,
    validate_and_split_simulation,
    validate_generation,
)
from deltakit_explorer._utils._logging import Logging
from deltakit_explorer.data._data_analysis import has_leakage
from deltakit_explorer.qpu._noise import SI1000Noise
from deltakit_explorer.types._experiment_types import (
    QECExperiment,
    QECExperimentDefinition,
)
from deltakit_explorer.types._types import (
    BinaryDataType,
    Decoder,
    DecodingResult,
    DetectionEvents,
    LeakageFlags,
    Measurements,
    ObservableFlips,
    PhysicalNoiseModel,
    QubitCoordinateToDetectorMapping,
    SI1000NoiseModel,
)


# pylint: disable=too-many-locals,unsubscriptable-object
class Client:
    """The `Client` class provides convenient methods to access circuit generation,
    noise addition, decoding, simulation, and analysis calls. To create a connection
    to a particular server, use the following syntax::

        client = Client(
            base_url="https://deltakit.riverlane.com/proxy"
        )

    """

    def __init__(
        self,
        base_url: str,
        timeout: int = 60000,
        api_version: int = 2,
    ):
        """
        Args:
          base_url : str
              Url of the Dektakit service endpoint.
          timeout: int
              Request timeout, seconds.
          api_version: int
              1 - Long living GraphQL requests.
              2 - Task-based requests.
        """
        if api_version == 1:
            self._api: APIClient = GQLClient(base_url, timeout)
        elif api_version == 2:
            self._api = APIv2Client(base_url, timeout)
        else:
            msg = f"API version {api_version} is not implemented."
            raise NotImplementedError(msg)
        self._api_version = api_version

    @staticmethod
    def get_instance(api_version: int=2) -> Client:
        """Return a pre-configured instance of a cloud-based Deltakit client.
        If you need to connect to a custom instance, please use `Client(base_url)`
        syntax.

        Args:
            api_version (int):
                version of API, GraphQL-based (1),
                or queue-based (2)

        Returns:
            Client: configured instance of a Deltakit API.
        """
        utils.load_environment_variables_from_drive()
        server_name = os.environ.get(
            utils.DELTAKIT_SERVER_URL_ENV,
            default=utils.DELTAKIT_SERVER_DEFAULT_URL_ENV,
        )
        return Client(base_url=server_name, api_version=api_version)

    @classmethod
    def set_token(cls, token: str, validate: bool = True):
        """Set and validate a token."""
        # NB: .get_instance() call fails if there is no token provided
        # as an env variable, or on a drive. For cold start, we supply
        # the variable first:
        os.environ[TOKEN_VARIABLE] = token
        return cls.get_instance()._api.set_token(token, validate)

    def add_noise(
        self,
        stim_circuit: str | stim.Circuit,
        noise_model: PhysicalNoiseModel | SI1000NoiseModel | SI1000Noise,
    ) -> str:
        """Given a stim circuit, changes all noise in the circuit to noise
        defined by user-given parameters. If the circuit is noiseless,
        noise is added into the circuit.

        Args:
            stim_circuit: (str | stim.Circuit):
                Noiseless circuit.
            noise_model (PhysicalNoiseModel | SI1000NoiseModel | SI1000Noise):
                Noise model to apply to a circuit.

        Returns:
            str:
                STIM circuit with error mechanisms derived from the
                noise model. Note, that adding SI1000NoiseModel may
                add leakage terms, which are not compatible with vanilla
                STIM, and can only be simulated using Deltakit client.

        Examples:
            Populate the circuit with the noise, and then simulate::

                leakage_noise_model = types.SI1000NoiseModel(p=0.001, p_l=0.001)
                noisy_circuit = client.add_noise(
                    stim_circuit=compiled_circuit,
                    noise_model=leakage_noise_model,
                )
                measurements, leakage = client.simulate_stim_circuit(
                    stim_circuit=noisy_circuit,
                    shots=num_shots,
                )

        """
        query_id = Logging.info_and_generate_uid(locals())
        try:
            if isinstance(noise_model, SI1000Noise):
                noise_model = SI1000NoiseModel(p=noise_model.p, p_l=noise_model.pL)
            return self._api.add_noise(stim_circuit, noise_model, query_id)
        except Exception as ex:
            Logging.error(ex, query_id)
            raise

    @validate_and_split_decoding
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def decode(
        self,
        detectors: DetectionEvents,
        observables: ObservableFlips,
        decoder: Decoder,
        noisy_stim_circuit: str | stim.Circuit,
        leakage_flags: LeakageFlags | None = None,
    ) -> DecodingResult:
        """Accepts detectors, observables and a stim circuit with integrated noise
        and returns decoding results for a requested decoder.

        Args:
            detectors (DetectionEvents):
                Syndrome data object.
            observables (ObservableFlips):
                Observables data.
            decoder (Decoder):
                Definition of decoder and its features.

                - `MWPM` for Minimum-Weight Perfect Matching
                  (https://arxiv.org/abs/2303.15933),
                - `CC` for Collision Clustering (https://arxiv.org/abs/2309.05558), or
                - `BELIEF_MATCHING` for Belief Matching
                  (https://arxiv.org/abs/2203.04948).
                - `BP_OSD` for Belief Propagation - Ordered Statistics Decoding
                  (BP-OSD) (https://quantum-journal.org/papers/q-2021-11-22-585/),
                - `AC` for Ambiguity Clustering (https://arxiv.org/abs/2406.14527).
                - `LCD` for Local Clustering Decoder (https://arxiv.org/abs/2411.10343).

            noisy_stim_circuit (str | stim.Circuit):
                STIM circuit with defined noise terms. Used to derive detector
                error model.

            leakage_flags (Optional[LeakageFlags]):
                Leakage information may be used by Local Clustering Decoder (LCD) to
                improve decoding quality.

        Returns:
            DecodingResult: Server decoding response.

        Examples:
            Decoding simulated data::

                measurements, leakage = client.simulate_stim_circuit(
                    stim_circuit=noisy_circuit,
                    shots=10000,
                )
                detectors, observables = measurements.to_detectors_and_observables(
                    stim_circuit=noisy_circuit,
                )
                decoding_result = client.decode(
                    detectors=detectors,
                    observables=observables,
                    decoder=types.Decoder(types.DecoderType.AC),
                    noisy_stim_circuit=noisy_circuit,
                )

        """
        query_id = Logging.info_and_generate_uid(locals())
        try:
            return self._api.decode(
                detectors,
                observables,
                decoder,
                noisy_stim_circuit,
                leakage_flags,
                query_id,
            )
        except Exception as ex:
            Logging.error(ex, query_id)
            raise

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def decode_measurements(
        self,
        measurements: Measurements,
        decoder: Decoder,
        ideal_stim_circuit: str | stim.Circuit,
        noise_model: PhysicalNoiseModel,
        leakage_flags: LeakageFlags | None = None,
        sweep_bits: BinaryDataType | None = None,
    ) -> DecodingResult:
        """Accept raw measurement file and a clean stim circuit and
        return decoding results with a requested decoder.

        Args:
            measurements (Measurements):
                Measurements object.
            decoder (Decoder):
                Predefined decoder object.

                - `MWPM` for Minimum-Weight Perfect Matching
                  (https://arxiv.org/abs/2303.15933),
                - `CC` for Collision Clustering (https://arxiv.org/abs/2309.05558), or
                - `BELIEF_MATCHING` for Belief Matching
                  (https://arxiv.org/abs/2203.04948).
                - `BP_OSD` for Belief Propagation - Ordered Statistics Decoding
                  (BP-OSD) (https://quantum-journal.org/papers/q-2021-11-22-585/),
                - `AC` for Ambiguity Clustering (https://arxiv.org/abs/2406.14527).
                - `LCD` for Local Clustering Decoder (https://arxiv.org/abs/2411.10343).

            ideal_stim_circuit: (str | stim.Circuit): clean stim circuit.
            noise_model (PhysicalNoiseModel):
                Noise model used to generate a circuit, or a noise model close
                to hardware, which has generated the measurements data.
            leakage_flags (Optional[LeakageFlags]):
                Heralded leakage events, if available.
            sweep_bits (Optional[BinaryDataType]):
                Initialisation bits, if were used in experiments.

        Returns:
            DecodingResult: server response.

        Examples:
            Decoding using experimental graph method::

                decoder = Decoder(
                    decoder_type=DecoderType.MWPM,
                    use_experimental_graph=True
                )
                client.decode_measurements(
                    measurements=measurements,
                    decoder=decoder,
                    ideal_stim_circuit=circuit,
                    noise_model=PhysicalNoiseModel.get_floor_superconducting_noise(),
                    sweep_bits=sweep_bits,
                )

            Decoding with AC with additional parameters::

                decoder = types.Decoder(
                    decoder_type=types.DecoderType.AC,
                    parallel_jobs=8,
                    parameters={
                        "decompose_errors": False,
                        "bp_rounds": 99,
                        "ac_kappa_proportion": 0.02
                    }
                )
                client.decode_measurements(
                    measurements=measurements,
                    decoder=decoder,
                    ideal_stim_circuit=compiled_circuit,
                    noise_model=noise_model,
                )

        """
        noisy_stim_circuit = self.add_noise(
            stim_circuit=ideal_stim_circuit,
            noise_model=noise_model,
        )
        noisy_stim_file_content = noisy_stim_circuit
        dets, obs = measurements.to_detectors_and_observables(
            stim_circuit=noisy_stim_file_content,
            sweep_bits=sweep_bits,
        )
        return self.decode(
            detectors=dets,
            observables=obs,
            decoder=decoder,
            noisy_stim_circuit=noisy_stim_circuit,
            leakage_flags=leakage_flags,
        )

    def defect_rates(
        self,
        detectors: DetectionEvents,
        stim_circuit: str | stim.Circuit,
    ) -> dict[tuple[float, ...], list[float]]:
        """Obtain defect rates for the detectors
        defined in the stim circuit.

        Args:
            detectors (DetectionEvents):
                Detectors data.
            stim_circuit (str | stim.Circuit):
                STIM circuit content.

        Returns:
            Dict[Tuple[float, ...], List[float]]:
                Keys are detectors coordinates, values are
                their defect rates.

        Examples:
            Obtaining a defect rate list for a single coordinate::

                coordinate = (2.0, 5.0)
                defect_rates = client.defect_rates(detectors, stim_circuit)
                print(f"{coordinate}: {defect_rates[coordinate]}")

        """
        query_id = Logging.info_and_generate_uid(locals())
        try:
            return self._api.defect_rates(
                detectors, stim_circuit, query_id,
            )
        except Exception as ex:
            Logging.error(ex, query_id)
            raise

    def get_correlation_matrix_for_trimmed_data(
        self,
        detectors: DetectionEvents,
        noise_floor_circuit: str | stim.Circuit,
        use_default_noise_model_edges: bool = False,
    ) -> tuple[npt.NDArray[np.float64], QubitCoordinateToDetectorMapping]:
        """Obtain a correlation matrix for a given set of detection events
        and corresponding noisy circuit. Also returns a qubit-detector mapping
        that can be used to plot detectors in groups according to the qubit
        they correspond to. This method may be used separately, but we recommend
        considering the simpler method Client.get_correlation_matrix, which calls it.

        Args:
            detectors (DetectionEvents):
                Detectors data.
            noise_floor_circuit (str | stim.Circuit):
                Cricuit with defined minimal noise. For example, this noise may
                be obtained with
                PhysicalNoiseModel.get_floor_superconducting_noise().
            use_default_noise_model_edges (bool):
                If set to True, uses noise edges defined in the circuit, otherwise
                derives noise from detectors data.

        Returns:
            Tuple[npt.NDArray, QubitCoordinateToDetectorMapping]:
                A matrix with correlation values,
                and a mapping of coordinates to detectors.

        """
        query_id = Logging.info_and_generate_uid(locals())
        try:
            return self._api.get_correlation_matrix_for_trimmed_data(
                detectors, noise_floor_circuit, use_default_noise_model_edges, query_id,
            )
        except Exception as ex:
            Logging.error(ex, query_id)
            raise

    def get_correlation_matrix(
        self,
        detectors: DetectionEvents,
        stim_circuit: str | stim.Circuit,
        use_default_noise_model_edges: bool,
    ) -> tuple[npt.NDArray[np.float64], QubitCoordinateToDetectorMapping]:
        """Obtain a correlation matrix for a given set of detection events
        and corresponding circuit. Also returns a qubit-detector mapping
        that can be used to plot detectors in groups according to the qubit
        they correspond to. First trims the circuit and detector data.

        Args:
            detectors (DetectionEvents):
                Detectors data.
            stim_circuit (str | stim.Circuit):
                Clean STIM circuit.
            use_default_noise_model_edges (bool):
                If set to True, uses noise edges defined in the circuit, otherwise
                derives noise from detectors data.

        Returns:
            Tuple[npt.NDArray, QubitCoordinateToDetectorMapping]:
                - The correlation matrix for the given data;
                - Mapping of a qubit (in coordinates) to the detectors it
                  corresponds to.

        Examples:
            Getting the matrix and plotting it::

                matrix, mapping = client.get_correlation_matrix(
                    detectors, stim_circuit,
                    use_default_noise_model_edges=True,
                )
                plt = visualisation.correlation_matrix(matrix, mapping)

        """
        query_id = Logging.info_and_generate_uid(locals())
        try:
            noisy_circuit = self.add_noise(
                stim_circuit=stim_circuit,
                noise_model=PhysicalNoiseModel.get_floor_superconducting_noise(),
            )
            (
                trimmed_circuit,
                trimmed_dets
            ) = self.trim_circuit_and_detectors(noisy_circuit, detectors)
            return self.get_correlation_matrix_for_trimmed_data(
                trimmed_dets, trimmed_circuit, use_default_noise_model_edges)
        except Exception as ex:
            Logging.error(ex, query_id)
            raise

    def trim_circuit_and_detectors(
        self,
        stim_circuit: str | stim.Circuit,
        detectors: DetectionEvents,
    ) -> tuple[str, DetectionEvents]:
        """Trims a circuit to remove redundant detectors (detectors that do not
        affect the observable). It then compares the trimmed circuit to the
        original to make a list of the removable detectors that are then used
        to also trim the detector data.

        Trimming a circuit:
            removing detectors that are not associated
            with any obervables.
        Trimming detector data:
            having removed detectors from the circuit
            as per the above definition, remove those same detectors
            from the given detector data.

        Args:
            stim_circuit (str | stim.Circuit):
                Noisy STIM circuit.
            detectors (DetectionEvents):
                Corresponding detectors.


        Returns:
            Tuple[str, DetectionEvents]
                Tuple containing trimmed circuit and
                detection events.

        Examples:
            Trim a disconnected circuit and related detection events,
            keeping only the part of it, connected to observables::

                trimmed_circuit, trimmed_dets = client.trim_circuit_and_detectors(
                    stim_circuit=noisy_circuit,
                    detectors=detectors,
                )


        """
        query_id = Logging.info_and_generate_uid(locals())
        try:
            return self._api.trim_circuit_and_detectors(
                stim_circuit, detectors, query_id,
            )
        except Exception as ex:
            Logging.error(ex, query_id)
            raise

    def get_experiment_detectors_and_defect_rates(
        self,
        experiment: QECExperiment,
    ) -> tuple[dict[int, tuple[float, ...]], dict[tuple[float, ...], list[float]]]:
        """Extract detector to coordinate mapping,
        and defect rates for these detectors.

        Args:
            experiment (QECExperiment):
                Experiment with measurements or detectors defined.

        Returns:
            Tuple[Dict[int, Tuple[float, ...]], Dict[Tuple[float, ...], List[float]]]:
                Two dictionaries: the first is a mapping from
                detector index to coordinates, the second is mapping
                from coordinates to defect rates per round.

        Examples:
            Defining experiment and extracting defect rate information::

                experiment = QECExperiment.from_circuit_and_measurements(
                    folder / "circuit_noisy.stim",
                    folder / "measurements.b8",
                    DataFormat.B8,
                )
                mapping, rates = client.get_experiment_detectors_and_defect_rates(
                    experiment
                )
        """
        if experiment.measurements is None and experiment.detectors is None:
            msg = (
                "Experiment object should have at least measurements or detectors. "
                "Provided object has neither."
            )
            raise ValueError(
                msg
            )
        experiment.compute_detectors_and_observables()
        # trim the circuit to remove redundant detectors
        trimmed_circuit, trimmed_dets = self.trim_circuit_and_detectors(
            stim_circuit=experiment.noisy_circuit,
            # detectors are computed above
            detectors=experiment.detectors,
        )
        # get defect rates for trimmed circuit
        all_qubit_defect_rates = self.defect_rates(
            detectors=trimmed_dets,
            stim_circuit=trimmed_circuit,
        )
        circuit = stim.Circuit(trimmed_circuit)
        all_detector_coordinates = {
            key: tuple(value) for key, value in
            circuit.get_detector_coordinates().items()
        }
        return all_detector_coordinates, all_qubit_defect_rates

    @validate_generation
    def generate_circuit(
        self,
        experiment_definition: QECExperimentDefinition,
    ) -> str:
        """Generate a STIM circuit for a quantum error correction experiment.
        `experiment_definition` holds all information essential for building
        an experiment.

        Args:
            experiment_definition (QECExperimentDefinition):
                Definition of an experiment.

                - experiment_type: Type of experiment, e.g. Quantum Memory or Stability.
                - code_type: QEC code, e.g. Rotated Planar Code.
                - observable_basis: Pauli Basis, in which observable is measured.
                  E.g. X or Z.
                - num_rounds: Number of experiment rounds.
                - basis_gates: If circuit is generated with a specific basis gate set.
                - parameters: Optional parameters of circuit generation.

        Returns:
            str: STIM circuit.

        Examples:

            Using shortcut experiment definition::

                compiled_circuit = client.generate_circuit(
                    QECExperimentDefinition.get_rotated_planar_z_quantum_memory(
                        distance, distance, ["CZ", "H", "MZ", "RZ"]
                    )
                )

            Using explicit definition::

                circuit = client.generate_circuit(
                    types.QECExperimentDefinition(
                        experiment_type=types.QECExperimentType.QUANTUM_MEMORY,
                        code_type=types.QECECodeType.BIVARIATE_BICYCLE,
                        observable_basis=PauliBasis.Z,
                        num_rounds=6,
                        basis_gates=["CZ", "H", "MZ", "RZ"],
                        parameters=types.CircuitParameters.from_matrix_specification(
                            param_l=6,
                            param_m=6,
                            m_A_powers=[3, 1, 2],
                            m_B_powers=[3, 1, 2],
                        )
                    )
                )

        """
        query_id = Logging.info_and_generate_uid(locals())
        try:
            return self._api.generate_circuit(experiment_definition, query_id)
        except Exception as ex:
            Logging.error(ex, query_id)
            raise

    @validate_and_split_simulation
    def simulate_stim_circuit(
        self,
        stim_circuit: str | stim.Circuit,
        shots: int,
    ) -> tuple[Measurements, LeakageFlags | None]:
        """
        Simulate STIM circuit with Deltakit client.

        Args:
            stim_circuit (str | stim.Circuit):
                Any STIM circuit. May include leakage instructions.
            shots (int):
                Number of shots.

        Returns:
            Tuple[Measurements, Optional[LeakageFlags]]:
                (Measurements, Leakage). Leakage may be None.

        Examples:
            Generate and simulate a circuit with leakage::

                compiled_circuit = client.generate_circuit(
                    types.QECExperimentDefinition.get_rotated_planar_z_quantum_memory(
                        distance=distance,
                        num_rounds=num_rounds,
                        basis_gates=["CZ", "H", "MZ", "RZ"],
                    )
                )
                # Add leakage noise
                leakage_noise_model = types.SI1000NoiseModel(p=0.001, p_l=0.001)
                noisy_circuit = client.add_noise(
                    stim_circuit=compiled_circuit,
                    noise_model=leakage_noise_model,
                )
                measurements, leakage = client.simulate_stim_circuit(
                    stim_circuit=noisy_circuit,
                    shots=num_shots,
                )

        """
        query_id = Logging.info_and_generate_uid(locals())
        try:
            # always simulate pure stim with no leakage locally
            if not has_leakage(str(stim_circuit)):
                return simulation.simulate_with_stim(stim_circuit, shots)
            return self._api.simulate_circuit(stim_circuit, shots, query_id)
        except Exception as ex:
            Logging.error(ex, query_id)
            raise

    def kill(self, request_id: str) -> int:
        """Kill a decoding task submitted to a client.
        Decoding methods report a request ID. Calling this
        method with an ID tries to kill all CPU
        workers associated with the task. Method returns the number
        of CPU workers it could terminate. If the task is already finished,
        the method should return 0.

        Args:
            request_id (str):
                ID of the requests which started the decoding task.

        Returns:
            int: number of workers killed
        """
        return self._api.kill(request_id)



# (c) Copyright Riverlane 2020-2025.
"""
This module stores an implementation of a function to construct a
circuit for a quantum memory experiment with a Calderbank-Shor-Steane
(CSS) quantum error correction code.
"""

from __future__ import annotations

import stim
from deltakit_circuit import Circuit
from deltakit_circuit.gates import PauliBasis

import deltakit_explorer
from deltakit_explorer.codes._bivariate_bicycle_code import BivariateBicycleCode
from deltakit_explorer.codes._css._experiment_circuit import experiment_circuit
from deltakit_explorer.codes._css._stabiliser_code import StabiliserCode
from deltakit_explorer.codes._planar_code._planar_code import PlanarCode
from deltakit_explorer.codes._planar_code._rotated_planar_code import RotatedPlanarCode
from deltakit_explorer.codes._planar_code._unrotated_planar_code import (
    UnrotatedPlanarCode,
)
from deltakit_explorer.codes._planar_code._unrotated_toric_code import (
    UnrotatedToricCode,
)
from deltakit_explorer.codes._repetition_code import RepetitionCode
from deltakit_explorer.enums import QECECodeType, QECExperimentType
from deltakit_explorer.types import CircuitParameters, QECExperimentDefinition


def css_code_memory_circuit(
    css_code: StabiliserCode,
    num_rounds: int,
    logical_basis: PauliBasis,
    client: deltakit_explorer.Client | None = None,
    use_iswap_gates: bool = False,
) -> Circuit:
    """
    Return a noiseless `deltakit.circuit.Circuit` for an X or Z quantum memory experiment
    for CSS codes.

    Parameters
    ----------
    css_code : StabiliserCode
        Stabiliser code.
    num_rounds : int
        Number of rounds to measure the stabilisers for.
    logical_basis : PauliBasis
        PauliBasis instance specifying whether to perform an X or Z quantum memory
        experiment.
    client : Optional[deltakit_explorer.Client]
        The `client` used to perform the calculation.
    use_iswap_gates : bool
        If you generate using cloud, you may request using
        iSWAP native gate set. Default is False.

    Returns
    -------
    Circuit
        Noiseless circuit for quantum memory.

    Raises
    ------
    ValueError
        If num_rounds is not positive.
    ValueError
        If logical_basis is neither PauliBasis.X nor PauliBasis.Z.
    """
    if num_rounds < 1:
        msg = "Invalid num_rounds, it has to be positive."
        raise ValueError(msg)
    if logical_basis not in [PauliBasis.X, PauliBasis.Z]:
        msg = "Invalid logical_basis, it has to be PauliBasis.X or PauliBasis.Z"
        raise ValueError(msg)
    if use_iswap_gates and client is None:
        msg = "`use_iswap_gates == True` is only supported when a `client` object is provided."
        raise NotImplementedError(msg)
    if client is not None:
        return _cloud_css_code_experiment_circuit(QECExperimentType.QUANTUM_MEMORY,
                                              css_code, num_rounds, logical_basis, client, use_iswap_gates)
    data_qubit_init_stage = (
        css_code.encode_logical_zeroes()
        if logical_basis == PauliBasis.Z
        else css_code.encode_logical_pluses()
    )
    stabiliser_meas_stage = css_code.measure_stabilisers(num_rounds=num_rounds)
    data_qubit_meas_stage = (
        css_code.measure_z_logicals()
        if logical_basis == PauliBasis.Z
        else css_code.measure_x_logicals()
    )
    return experiment_circuit(
        experiment=[data_qubit_init_stage, stabiliser_meas_stage, data_qubit_meas_stage]
    )


def _cloud_css_code_experiment_circuit(
    experiment_type: QECExperimentType,
    css_code: StabiliserCode,
    num_rounds: int,
    logical_basis: PauliBasis,
    client: deltakit_explorer.Client | None = None,
    use_iswap_gates: bool = False,
) -> Circuit:
    """
    Return a noiseless `deltakit.circuit.Circuit` for an X or Z quantum stability experiment
    for CSS codes.

    Parameters
    ----------
    experiment_type: QECExperimentType
        Type of experiment.
    css_code : StabiliserCode
        Stabiliser code.
    num_rounds : int
        Number of rounds to measure the stabilisers for.
    logical_basis : PauliBasis
        PauliBasis instance specifying whether to perform an X or Z quantum memory
        experiment.
    client : Optional[deltakit_explorer.Client]
        The `client` used to perform the calculation.
    use_iswap_gates : bool
        If you generate using cloud, you may request using
        iSWAP native gate set. Default is False.

    Returns
    -------
    Circuit
        Noiseless circuit for a quantum stability experiment.

    Raises
    ------
    NotImplementedError :
        If `client` is not provided.
    ValueError :
        If `css_code` is not of a valid type.
        If `use_iswap_gates` is used without `client`.
    """
    if use_iswap_gates and client is None:
        msg = "`use_iswap_gates == True` is only supported when a `client` object is provided."
        raise NotImplementedError(msg)
    if client is None:
        msg = "A `client` is required to obtain a stability circuit."
        raise NotImplementedError(msg)

    code_types = {
        RotatedPlanarCode: QECECodeType.ROTATED_PLANAR,
        UnrotatedPlanarCode: QECECodeType.UNROTATED_PLANAR,
        UnrotatedToricCode: QECECodeType.UNROTATED_TORIC,
        RepetitionCode: QECECodeType.REPETITION,
        BivariateBicycleCode: QECECodeType.BIVARIATE_BICYCLE,
    }
    code_type = code_types[css_code.__class__]

    if isinstance(css_code, PlanarCode):
        parameters = CircuitParameters.from_sizes(
            (css_code.width, css_code.height)
        )
    elif isinstance(css_code, RepetitionCode):
        parameters = CircuitParameters.from_sizes(
            (css_code.distance,)
        )
    elif isinstance(css_code, BivariateBicycleCode):
        parameters=CircuitParameters.from_matrix_specification(
            param_l=css_code.param_l,
            param_m=css_code.param_m,
            m_A_powers=css_code.m_A_powers,
            m_B_powers=css_code.m_B_powers,
        )
    else:
        msg = "Unrecognised `css_code` type."
        raise ValueError(msg)
    basis_gates: list[str] | None = None
    if use_iswap_gates:
        basis_gates = [
            "ISWAP",
            "MZ", "RZ",
            "X", "Y", "Z", "H",
            "S", "S_DAG", "SQRT_X", "SQRT_X_DAG",
        ]

    circuit = client.generate_circuit(
        QECExperimentDefinition(
            experiment_type=experiment_type,
            code_type=code_type,
            observable_basis=logical_basis,
            num_rounds=num_rounds,
            parameters=parameters,
            basis_gates=basis_gates,
        )
    )

    return Circuit.from_stim_circuit(stim.Circuit(circuit))


def css_code_stability_circuit(
    css_code: StabiliserCode,
    num_rounds: int,
    logical_basis: PauliBasis,
    client: deltakit_explorer.Client | None = None,
    use_iswap_gates: bool = False,
) -> Circuit:
    """
    Return a noiseless `deltakit.circuit.Circuit` for an X or Z quantum stability experiment
    for CSS codes.

    Parameters
    ----------
    css_code : StabiliserCode
        Stabiliser code.
    num_rounds : int
        Number of rounds to measure the stabilisers for.
    logical_basis : PauliBasis
        PauliBasis instance specifying whether to perform an X or Z quantum memory
        experiment.
    client : Optional[deltakit_explorer.Client]
        The `client` used to perform the calculation.
    use_iswap_gates : bool
        If you generate using cloud, you may request using
        iSWAP native gate set. Default is False.

    Returns
    -------
    Circuit
        Noiseless circuit for a quantum stability experiment.

    Raises
    ------
    NotImplementedError :
        If `client` is not provided.
    ValueError :
        If `css_code` is not of a valid type.
    """
    return _cloud_css_code_experiment_circuit(QECExperimentType.STABILITY,
                                              css_code, num_rounds, logical_basis, client, use_iswap_gates)



# (c) Copyright Riverlane 2020-2025.
"""
This module includes implementations of phenomenological
noise. `PhenomenologicalNoise` adds noise to `I` gates.
`ToyPhenomenologicalNoise` specifies this noise
to be Depolarise1, and adds measurement flip noise.
"""

from collections.abc import Callable
from dataclasses import dataclass

from deltakit_circuit import Qubit, measurement_noise_profile
from deltakit_circuit.gates import I
from deltakit_circuit.noise_channels import Depolarise1, OneQubitNoiseChannel

from deltakit_explorer.qpu._noise._noise_parameters import NoiseParameters


@dataclass(kw_only=True)
class PhenomenologicalNoise(NoiseParameters):
    """
    Class for capturing phenomenological noise.

    Parameters
    ----------
    phenomenological_noise : Optional[Callable[[Qubit], OneQubitNoiseChannel]]
        Phenomenological noise which depends on qubit only. By default,
        no noise.
    """

    phenomenological_noise: Callable[[Qubit], OneQubitNoiseChannel] | None = None

    def __post_init__(self):
        if self.phenomenological_noise is not None:
            self.gate_noise.append(
                lambda noise_context: [
                    self.phenomenological_noise(qubit)
                    for qubit in noise_context.gate_layer_qubits(I)
                ]
            )


@dataclass(kw_only=True)
class ToyPhenomenologicalNoise(PhenomenologicalNoise):
    """
    Class for capturing a phenomenological noise model with simple input parameters.

    Parameters
    ----------
    p : float, optional
        The depolarising error rate for phenomenological noise. By default, 0.0.
    p_measurement_flip: Optional[float]
        The probability of obtaining an incorrect measurement result. By default, this
        has the same value as p.
    """

    p: float = 0.0
    p_measurement_flip: float | None = None

    def __post_init__(self):
        if self.p_measurement_flip is None:
            self.p_measurement_flip = self.p
        self.measurement_flip = measurement_noise_profile(self.p_measurement_flip)
        self.phenomenological_noise = lambda qubit: Depolarise1(qubit, self.p)
        super().__post_init__()



# (c) Copyright Riverlane 2020-2025.
"""Provide functions to perform error-budgeting estimations."""

from ._budget import get_error_budget
from ._discretisation import (
    DiscretisationStrategy,
    get_linear_points,
    get_logarithmic_points,
)
from ._generation import generate_decoder_managers_for_lambda
from ._gradient import inverse_lambda_gradient_at
from ._lambda import inverse_lambda_at
from ._memory import MemoryGenerator, get_rotated_surface_code_memory_circuit
from ._parameters import FittingParameters, SamplingParameters
from ._post_processing import compute_lambda_and_stddev_from_results
from ._visualisation import plot_error_budget



# (c) Copyright Riverlane 2020-2025.
"""
This module includes implementation of the unrotated toric code. The code
represents two logical qubits.
"""

import itertools
from pathlib import Path

import matplotlib.pyplot as plt
import numpy.typing as npt
from deltakit_circuit import PauliX, PauliZ, Qubit
from deltakit_circuit._basic_types import Coord2D, Coord2DDelta
from deltakit_circuit._qubit_identifiers import PauliGate
from matplotlib.lines import Line2D
from matplotlib.patches import Patch
from numpy import arctan2, argsort, array, int8

from deltakit_explorer.codes._planar_code._planar_code import PlanarCode, ScheduleType
from deltakit_explorer.codes._schedules._schedule_order import (
    ScheduleOrder,
    get_x_and_z_schedules,
)
from deltakit_explorer.codes._schedules._unrotated_planar_code_schedules import (
    UnrotatedPlanarCodeSchedules,
)
from deltakit_explorer.codes._stabiliser import Stabiliser
from deltakit_explorer.enums._basic_enums import DrawingColours


class UnrotatedToricCode(PlanarCode):
    """
    Class representing the default unrotated toric code. The code has a periodic
    boundary and encodes two logical qubits. Logical operators are formed of loops around
    the torus. See the graph below showing a default 3x3 toric code. Open edges are
    connected to qubits on opposite sides. The distance of the code is given by the
    min(horizontal_distance, vertical_distance).

    This class also contains methods that help set up circuits for memory
    experiments.


    .. code-block:: text

        6├
        |    |      |      |      |      |      |
        5├    X ---- o ---- X ---- o ---- X ---- o ----
        |    |      |      |      |      |      |
        4├    ○ ---- Z ---- ○ ---- Z ---- ○ ---- Z ----
        │    |      |      |      |      |      |
        3├    X ---- ○ ---- X ---- ○ ---- X ---- o ----
        │    |      |      |      |      |      |
        2├    ○ ---- Z ---- ○ ---- Z ---- ○ ---- Z ----
        │    |      |      |      |      |      |
        1├    X ---- ○ ---- X ---- ○ ---- X ---- o ----
        │    |      |      |      |      |      |
        0├    ○ ---- Z ---- ○ ---- Z ---- ○ ---- Z ----
        │
        └----┴------┴------┴------┴------┴------┴------┴
            0      1      2      3      4      5      6

    Parameters
    ----------
    horizontal_distance: int
        The width of the toric code patch, which defines the distance for the
        horizontal logical operators X1 and Z2.
    vertical_distance: int
        The height of the toric code patch, which defines the distance for the
        vertical logical operators X2 and Z1.
    """

    def __init__(
        self,
        horizontal_distance: int,
        vertical_distance: int,
        schedule_type: ScheduleType = ScheduleType.SIMULTANEOUS,
        schedule_order: ScheduleOrder = ScheduleOrder.STANDARD,
        use_ancilla_qubits: bool = True,
    ):
        x_schedule, z_schedule = get_x_and_z_schedules(
            UnrotatedPlanarCodeSchedules, schedule_order
        )

        self._perform_css_checks = False

        super().__init__(
            width=horizontal_distance,
            height=vertical_distance,
            untransformed_x_schedule=x_schedule,
            untransformed_z_schedule=z_schedule,
            schedule_type=schedule_type,
            use_ancilla_qubits=use_ancilla_qubits,
        )

    def _calculate_untransformed_all_qubits(
        self,
    ) -> tuple[set[Qubit], set[Qubit], set[Qubit]]:
        data_qubits = {
            Qubit(Coord2D(x, y))
            for (x, y) in itertools.product(
                range(0, 2 * self.width - 1, 2),
                range(0, 2 * self.height - 1, 2),
            )
        }.union(
            {
                Qubit(Coord2D(x, y))
                for (x, y) in itertools.product(
                    range(1, 2 * self.width, 2),
                    range(1, 2 * self.height, 2),
                )
            }
        )

        x_ancilla_qubits = {
            Qubit(Coord2D(x, y))
            for (x, y) in itertools.product(
                range(0, 2 * self.width - 1, 2),
                range(1, 2 * self.height, 2),
            )
        }

        z_ancilla_qubits = {
            Qubit(Coord2D(x, y))
            for (x, y) in itertools.product(
                range(1, 2 * self.width, 2),
                range(0, 2 * self.height - 1, 2),
            )
        }

        return x_ancilla_qubits, z_ancilla_qubits, data_qubits

    def _calculate_single_type_stabilisers(
        self,
        ancilla_qubits: set[Qubit],
        schedule: tuple[Coord2DDelta, ...],
        gate: type[PauliGate],
    ) -> tuple[Stabiliser, ...]:
        stabilisers = []
        for ancilla in ancilla_qubits:
            paulis = []
            for delta in schedule:
                coordinate = Coord2D(*ancilla.unique_identifier) + delta
                x_mod = coordinate[0] % (2 * self.width)
                z_mod = coordinate[1] % (2 * self.height)
                qubit = Qubit(Coord2D(x_mod, z_mod))
                paulis.append(gate(qubit))

            stabilisers.append(Stabiliser(paulis=paulis, ancilla_qubit=ancilla))

        return tuple(stabilisers)

    def _calculate_untransformed_logical_operators(
        self,
    ) -> tuple[tuple[set[PauliGate], ...], tuple[set[PauliGate], ...]]:
        x_logicals = (
            {PauliX(Qubit(Coord2D(x, 0))) for x in range(0, 2 * self.width - 1, 2)},
            {PauliX(Qubit(Coord2D(1, y))) for y in range(1, 2 * self.height, 2)},
        )
        z_logicals = (
            {PauliZ(Qubit(Coord2D(0, y))) for y in range(0, 2 * self.height - 1, 2)},
            {PauliZ(Qubit(Coord2D(x, 1))) for x in range(1, 2 * self.width, 2)},
        )

        return (x_logicals, z_logicals)

    def draw_patch(self, filename: str | None = None, unrotated_code: bool = False) -> None:  # noqa: ARG002
        fig, ax = plt.subplots(nrows=1, ncols=1)
        all_qubit_x_coords = [qubit.unique_identifier.x for qubit in self.qubits]
        all_qubit_y_coords = [qubit.unique_identifier.y for qubit in self.qubits]
        min_x, max_x = min(all_qubit_x_coords) - 2, max(all_qubit_x_coords) + 2
        min_y, max_y = min(all_qubit_y_coords) - 2, max(all_qubit_y_coords) + 2
        ax.set_xlim((min_x, max_x))
        ax.set_ylim((min_y, max_y))

        stabilisers: tuple[Stabiliser, ...] = tuple(
            itertools.chain.from_iterable(self._stabilisers)
        )

        # Draw stabiliser plaquettes
        for stabiliser in stabilisers:
            data_qubit_x_coords = [
                pauli.qubit.unique_identifier[0]
                for pauli in stabiliser.paulis
                if pauli is not None
            ]

            data_qubit_y_coords = [
                pauli.qubit.unique_identifier[1]
                for pauli in stabiliser.paulis
                if pauli is not None
            ]

            # Wrap boundary stabilisers on the X axis
            if 0 in data_qubit_x_coords and 2 * self.width - 1 in data_qubit_x_coords:
                if data_qubit_x_coords.count(0) > data_qubit_x_coords.count(
                    2 * self.width - 1
                ):
                    data_qubit_x_coords = [
                        -1 if x == 2 * self.width - 1 else x
                        for x in data_qubit_x_coords
                    ]
                else:
                    data_qubit_x_coords = [
                        2 * self.width if x == 0 else x for x in data_qubit_x_coords
                    ]

            # Wrap boundary stabilisers on the Y axis
            if 0 in data_qubit_y_coords and 2 * self.height - 1 in data_qubit_y_coords:
                if data_qubit_y_coords.count(0) > data_qubit_y_coords.count(
                    2 * self.height - 1
                ):
                    data_qubit_y_coords = [
                        -1 if y == 2 * self.height - 1 else y
                        for y in data_qubit_y_coords
                    ]
                else:
                    data_qubit_y_coords = [
                        2 * self.height if y == 0 else y for y in data_qubit_y_coords
                    ]

            # Re order data qubit coords for polygon drawing
            ordered_data_qubit_y_coords: npt.NDArray[int8] = array(data_qubit_y_coords)
            ordered_data_qubit_x_coords: npt.NDArray[int8] = array(data_qubit_x_coords)

            order = argsort(
                arctan2(
                    ordered_data_qubit_y_coords - ordered_data_qubit_y_coords.mean(),
                    ordered_data_qubit_x_coords - ordered_data_qubit_x_coords.mean(),
                )
            )

            paulis = [pauli for pauli in stabiliser.paulis if pauli is not None]
            ax.fill(
                ordered_data_qubit_x_coords[order],
                ordered_data_qubit_y_coords[order],
                color=(
                    DrawingColours.X_COLOUR.value
                    if isinstance(paulis[0], PauliX)
                    else DrawingColours.Z_COLOUR.value
                ),
                alpha=1,
            )

        # Draw data qubits
        for qubit in self._data_qubits:
            cc = plt.Circle(
                qubit.unique_identifier,
                0.2,
                color=DrawingColours.DATA_QUBIT_COLOUR.value,
                alpha=1,
            )
            ax.set_aspect(1)
            ax.add_artist(cc)

        # Draw stabiliser ancilla qubits
        if self._use_ancilla_qubits:
            # Draw X stabiliser ancilla qubits
            for qubit in self._x_ancilla_qubits:
                cc = plt.Circle(
                    qubit.unique_identifier,
                    0.2,
                    color=DrawingColours.ANCILLA_QUBIT_COLOUR.value,
                    alpha=1,
                )
                ax.set_aspect(1)
                ax.add_artist(cc)

            # Draw X stabiliser ancilla qubits
            for qubit in self._z_ancilla_qubits:
                cc = plt.Circle(
                    qubit.unique_identifier,
                    0.2,
                    color=DrawingColours.ANCILLA_QUBIT_COLOUR.value,
                    alpha=1,
                )
                ax.set_aspect(1)
                ax.add_artist(cc)

        # Create a legend
        legend_elements = [
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                label="Data Qubit",
                markerfacecolor=DrawingColours.DATA_QUBIT_COLOUR.value,
                markersize=15,
            ),
            Line2D(
                [0],
                [0],
                marker="o",
                color="w",
                label="Ancilla Qubit",
                markerfacecolor=DrawingColours.ANCILLA_QUBIT_COLOUR.value,
                markersize=15,
            ),
            Patch(facecolor=DrawingColours.X_COLOUR.value, label="X Stabiliser"),
            Patch(facecolor=DrawingColours.Z_COLOUR.value, label="Z Stabiliser"),
        ]
        legend = ax.legend(
            handles=legend_elements,
            loc="upper center",
            bbox_to_anchor=(0.5, -0.1),
            ncol=2,
        )

        if filename is not None:
            # Save the file
            output_directory = Path(filename)
            if not output_directory.exists():
                output_directory.parent.mkdir(parents=True, exist_ok=True)
            fig.savefig(filename, bbox_extra_artists=(legend,), bbox_inches="tight")
            plt.close(fig)



# (c) Copyright Riverlane 2020-2025.
"""
This module contains an implementation of a QPU simulator class.
The class provides native gate compilation, noise addition and
execution time computation features. This is a fully-functional class.
"""
# pylint: disable=too-many-branches
from collections.abc import Iterable
from copy import deepcopy
from typing import NamedTuple

from deltakit_circuit import Circuit, GateLayer, NoiseLayer, Qubit
from deltakit_circuit.gates import I

from deltakit_explorer.qpu._circuits import (
    compile_circuit_to_native_gates,
    remove_identities,
)
from deltakit_explorer.qpu._native_gate_set import (
    ExhaustiveGateSet,
    NativeGateSetAndTimes,
)
from deltakit_explorer.qpu._noise import NoiseParameters, PhenomenologicalNoise


class CircuitSchedule(NamedTuple):
    active_times_list: list[dict[Qubit, float]]
    previous_layer_times: list[float]



class QPU:
    """
    Class capturing properties and functionality of a QPU. The class can
    compile a circuit to the device's native gates and add noise to a
    circuit corresponding to the device's noise model.

    Parameters
    ----------
    qubits : Set[Qubit]
        The device qubits.
    native_gates_and_times : NativeGateSetAndTimes
        The native gates (including measurement and reset) and the time they
        take to apply.
    noise_model : NoiseParameters
        The noise associated with operations on the QPU.
    maximise_parallelism : bool, optional
        Whether to do as many gates in parallel as possible (True) or only to
        perform gates in parallel if they are in the same GateLayer (False).
        By default, True.
    """

    def __init__(
        self,
        qubits: Iterable[Qubit],
        native_gates_and_times: NativeGateSetAndTimes = NativeGateSetAndTimes(),
        noise_model: NoiseParameters = NoiseParameters(),
        maximise_parallelism: bool = True,
    ):
        self.qubits = set(qubits)
        self.native_gates_and_times = native_gates_and_times
        self.noise_model = noise_model
        self._maximise_parallelism = maximise_parallelism

        if isinstance(noise_model, PhenomenologicalNoise):
            self.native_gates_and_times.add_gate(I, 0.0)

        self.full_native_gates_and_times = {
            **self.native_gates_and_times.one_qubit_gates,
            **self.native_gates_and_times.two_qubit_gates,
            **self.native_gates_and_times.reset_gates,
            **self.native_gates_and_times.measurement_gates,
        }

    def _get_circuit_schedule(self, circuit: Circuit) -> CircuitSchedule:
        if not circuit.qubits.issubset(self.qubits):
            qubit_ids = {q.unique_identifier for q in circuit.qubits - self.qubits}
            msg = f"Qubits {qubit_ids} in the circuit are not present on the QPU."
            raise ValueError(msg)

        nonnative_gateset = set()
        active_times_list, previous_layer_times = [], []
        active_times = dict.fromkeys(self.qubits, 0.0)
        previous_layer_time = 0.0

        for layer in circuit.layers:
            if not isinstance(layer, (GateLayer, Circuit)):
                # these numbers don't change with non-gate layers
                previous_layer_times.append(previous_layer_time)
                active_times_list.append(active_times.copy())
                continue
            if isinstance(layer, Circuit):
                # Finish previous layer before entering repeat block
                previous_layer_time = max(active_times.values())
            # Calculate previous layer time based on qubits acted upon in current layer
            elif self._maximise_parallelism:
                previous_layer_time = max(
                    (
                        time
                        for (qubit, time) in active_times.items()
                        if qubit in layer.qubits
                    ),
                    default=0,
                )
            else:
                previous_layer_time = max(
                    active_times.values(),
                    default=0,
                )

            # here we record values for noise computation
            # noise according these values will be added
            # before the layer.
            previous_layer_times.append(previous_layer_time)
            active_times_list.append(active_times.copy())

            # Update qubit active times
            active_times = {
                qubit: max(0.0, time - previous_layer_time)
                for (qubit, time) in active_times.items()
            }
            if isinstance(layer, GateLayer):
                # for a gate layer add active time to qubits
                for gate in layer.gates:
                    for qubit in gate.qubits:
                        # check that the gate is a native gate
                        if type(gate) not in self.full_native_gates_and_times:
                            nonnative_gateset.add(type(gate).__name__)
                        # add execution time to the gate
                        active_times[qubit] += self.full_native_gates_and_times[
                            type(gate)
                        ]
            elif isinstance(layer, Circuit):
                # NB: for a circuit layer add circuit time
                # to all qpu (!) qubits
                circuit_time = self.get_circuit_execution_time(layer)
                for qubit in self.qubits:
                    active_times[qubit] = circuit_time

        # the final layer
        previous_layer_time = max(active_times.values())
        previous_layer_times.append(previous_layer_time)
        active_times_list.append(active_times.copy())

        if nonnative_gateset:
            msg = (
                f"Gate(s) {nonnative_gateset} present in the circuit "
                "do not belong to the native gate set."
            )
            raise ValueError(msg)
        return CircuitSchedule(active_times_list, previous_layer_times)

    def _apply_idle_noise(self, circuit: Circuit) -> Circuit:
        """
        Add idle noise to the input circuit.

        Parameters
        ----------
        circuit : Circuit
            Circuit to which to add idle noise.

        Returns
        -------
        Circuit
            Circuit with idle noise added.
        """

        def _get_idle_noise_channels(
            previous_layer_time: float, active_times: dict[Qubit, float]
        ):
            if (
                previous_layer_time > 0.0
                and (idle_noise := self.noise_model.idle_noise) is not None
            ):
                idle_noise_channels = [
                    idle_noise(qubit, previous_layer_time - active_times[qubit])
                    for qubit in self.qubits
                    if previous_layer_time - active_times[qubit] > 0.0
                ]
            else:
                idle_noise_channels = []
            return idle_noise_channels

        layers_with_noise = []
        iterations = circuit.iterations
        qubit_times, layer_times = self._get_circuit_schedule(circuit)

        for layer, active_times, previous_layer_time in zip(
            circuit.layers, qubit_times, layer_times
        ):
            if not isinstance(layer, (GateLayer, Circuit)):
                layers_with_noise.append(layer)
                continue
            if isinstance(layer, Circuit):
                # Add layers from repeat block
                layer_to_add = self._apply_idle_noise(layer)
            else:
                layer_to_add = GateLayer(layer.gates)

            # Apply idle noise associated with previous layer
            idle_noise_channels = _get_idle_noise_channels(
                previous_layer_time, active_times
            )
            if len(idle_noise_channels) > 0:
                layers_with_noise.append(NoiseLayer(idle_noise_channels))

            # Append current layer and update previous layer
            layers_with_noise.append(layer_to_add)

        previous_layer_time, active_times = layer_times[-1], qubit_times[-1]
        idle_noise_channels = _get_idle_noise_channels(
            previous_layer_time, active_times
        )
        if len(idle_noise_channels) > 0:
            layers_with_noise.append(NoiseLayer(idle_noise_channels))

        return Circuit(layers_with_noise, iterations)

    def get_circuit_execution_time(self, circuit: Circuit) -> float:
        """
        Compute the total circuit execution time (in seconds), given a
        compiled circuit. This provides an estimate of the runtime of the
        circuit on a real QPU.

        Parameters
        ----------
        circuit : Circuit
            Circuit to be assessed.

        Returns
        -------
        float
            Time (in seconds) of total circuit execution.

        Raises
        ------
        ValueError
            If circuit contains non-native gates for the QPU.
            If circuit uses qubits not present on the QPU.
        """
        total_time = 0.0
        schedule = self._get_circuit_schedule(circuit)
        for previous_layer_time, layer in zip(
            schedule.previous_layer_times, circuit.layers
        ):
            if not isinstance(layer, (GateLayer, Circuit)):
                continue
            total_time += previous_layer_time
        # Apply time from final layer
        total_time += schedule.previous_layer_times[-1]
        return total_time * circuit.iterations

    def add_noise_to_circuit(self, circuit: Circuit) -> Circuit:
        """
        Add noise to the supplied circuit based on the QPU's noise model.

        Parameters
        ----------
        circuit : Circuit
            Noiseless circuit to which to add noise.

        Returns
        -------
        Circuit
            Circuit with noise added.
        """
        noisy_circuit = Circuit(deepcopy(circuit.layers))

        # apply measurement flip noise
        noisy_circuit.replace_gates(self.noise_model.measurement_flip)

        noisy_circuit.apply_gate_noise(
            self.noise_model.as_noise_profile_after_gate(), Circuit.LayerAdjacency.AFTER
        )

        # add idle noise
        noisy_circuit = self._apply_idle_noise(noisy_circuit)

        if len(self.noise_model.measurement_noise_before) != 0:
            # add noise which occurs before gates
            noisy_circuit.apply_gate_noise(
                self.noise_model.measurement_noise_before,
                Circuit.LayerAdjacency.BEFORE,
            )

        return noisy_circuit

    def _get_valid_qubit_mapping(self, circuit: Circuit) -> dict[Qubit, Qubit]:
        """
        Get a mapping from qubits in the supplied circuit to qubits in the QPU.
        Currently, this can return only the trivial mapping and thus will raise
        an error if any qubit present in the input circuit is not present in the
        QPU.

        Parameters
        ----------
        circuit : Circuit
            The circuit for which to obtain a qubit mapping.

        Returns
        -------
        Dict[Qubit, Qubit]
            A dictionary which gives a mapping from qubits in the input circuit
            to qubits in the QPU.
        """
        mapping = {}
        for qubit in circuit.qubits:
            if qubit not in self.qubits:
                msg = (
                    "A valid circuit-to-QPU qubit mapping could not be found "
                    f"because qubit {qubit} is not present in the QPU."
                )
                raise ValueError(msg)
            mapping[qubit] = qubit

        return mapping

    def compile_circuit(
        self,
        circuit: Circuit,
        remove_paulis: bool = True,
    ) -> Circuit:
        """
        Compile the supplied circuit so that it uses the QPU's native gates.

        Parameters
        ----------
        circuit : Circuit
            Circuit to be compiled.
        remove_paulis : bool, optional
            Whether to remove Pauli gates from the circuit. This should be set
            to False if, for example, there is a desire to prepare a specific
            initial state which requires Pauli gates. By default, True.

        Returns
        -------
        Circuit
            Compiled circuit.
        """
        return compile_circuit_to_native_gates(
            circuit, self.native_gates_and_times, up_to_paulis=remove_paulis
        )

    def compile_and_add_noise_to_circuit(
        self,
        circuit: Circuit,
        remove_paulis: bool = True,
    ) -> Circuit:
        """
        Compile the supplied circuit so that it uses the QPU's native gates
        and add noise based on the QPU's noise model.

        Parameters
        ----------
        circuit : Circuit
            Circuit to be compiled and to which to add noise.
        remove_paulis : bool, optional
            Whether to remove Pauli gates from the circuit. This should be set
            to False if, for example, there is a desire to prepare a specific
            initial state which requires Pauli gates. By default, True.

        Returns
        -------
        Circuit
            Compiled circuit with noise added.
        """
        return remove_identities(
            self.add_noise_to_circuit(
                self.compile_circuit(circuit, remove_paulis)
            )
        )

    @staticmethod
    def from_circuit_compile_and_add_noise(
        circuit: Circuit,
        native_gates_and_times: NativeGateSetAndTimes = ExhaustiveGateSet(),
        noise_model: NoiseParameters = NoiseParameters(),
        remove_paulis: bool = True,
        maximise_parallelism: bool = True,
    ) -> Circuit:
        """
        Compile the supplied circuit with a given noise model on an unconstrained QPU
        which, by default, supports all gates natively.

        Parameters
        ----------
        circuit : Circuit
            Circuit to be compiled and to which to add noise.
        native_gates_and_times : NativeGateSetAndTimes
            The native gates (including measurement and reset) and the time they
            take to apply.
        noise_model : NoiseParameters
            The noise associated with operations on the QPU.
        remove_paulis : bool, optional
            Whether to remove Pauli gates from the circuit. This should be set
            to False if, for example, there is a desire to prepare a specific
            initial state which requires Pauli gates. By default, True.
        maximise_parallelism : bool, optional
            Whether to do as many gates in parallel as possible (True) or only to
            perform gates in parallel if they are in the same GateLayer (False).
            By default, True.

        Returns
        -------
        Circuit
            Compiled circuit with noise added.
        """
        qpu = QPU(
            qubits=deepcopy(circuit.qubits),
            native_gates_and_times=native_gates_and_times,
            noise_model=noise_model,
            maximise_parallelism=maximise_parallelism,
        )
        return qpu.compile_and_add_noise_to_circuit(
            circuit=circuit,
            remove_paulis=remove_paulis,
        )



# (c) Copyright Riverlane 2020-2025.
"""
This module includes implementation of a standard depolarising noise model, SD6.
This noise model sets all error probabilities to a single value p.
This model can be used for theory and simulation but is not
representative of real-world experiments.
"""

from dataclasses import dataclass

from deltakit_circuit import after_reset_flip_probability, measurement_noise_profile
from deltakit_circuit.gates import TWO_QUBIT_GATES, OneQubitCliffordGate
from deltakit_circuit.noise_channels import Depolarise1, Depolarise2

from deltakit_explorer.qpu._noise._noise_parameters import NoiseParameters


@dataclass(kw_only=True)
class SD6Noise(NoiseParameters):
    """
    The standard depolarising noise model as seen in QEC literature, e.g.:
    Table 2 in https://arxiv.org/abs/2108.10457
    This model sets all one- and two-qubit depolarising error probabilities to p,
    as well as setting all measurement flip probabilities to p.
    This model can be used for theory and simulation but is not
    representative of other models seen in real hardware devices.

    Parameters
    ----------
    p: float, optional
        Rate of all depolarising errors and measurement flips. By default, 0.0.
    """

    name = "sd6_noise"

    p: float = 0

    def __post_init__(self):
        depolarise1_generator = Depolarise1.generator_from_prob(self.p)
        depolarise2_generator = Depolarise2.generator_from_prob(self.p)

        self.gate_noise.append(
            lambda noise_context: depolarise1_generator(
                noise_context.gate_layer_qubits(
                    (*tuple(TWO_QUBIT_GATES), OneQubitCliffordGate), gate_qubit_count=1
                )
            )
        )

        self.gate_noise.append(
            lambda noise_context: depolarise2_generator(
                noise_context.gate_layer_qubits(
                    tuple(TWO_QUBIT_GATES), gate_qubit_count=2
                )
            )
        )

        self.idle_noise = lambda qubit, _t=0.0: Depolarise1(
            qubit=qubit, probability=self.p
        )

        self.reset_noise = after_reset_flip_probability(self.p)

        self.measurement_flip = measurement_noise_profile(self.p)

    def __str__(self) -> str:
        return f"{self.name}_{self.p:.0e}"



# (c) Copyright Riverlane 2020-2025.
"""
This module defines a function to compute the logical operators associated
with a collection of stabilisers.
"""

from __future__ import annotations

from collections.abc import Collection, Iterable

import numpy as np
from deltakit_circuit import PauliX, PauliY, PauliZ, Qubit
from deltakit_circuit._qubit_identifiers import _PauliGate
from ldpc import mod2
from numpy.typing import NDArray
from stim import PauliString, Tableau

from deltakit_explorer.codes._css._stabiliser_helper_functions import (
    pauli_gates_to_stim_pauli_string,
)
from deltakit_explorer.codes._stabiliser import Stabiliser


def paulistring_to_operator(
    paulistr: PauliString, index_to_qubit: dict[int, Qubit]
) -> list[_PauliGate]:
    """
    Converts a stim PauliString to a list of PauliGate objects.

    Parameters
    ----------
    paulistr : stim.PauliString
        The stim pauli string.
    index_to_qubit : dict[int, Qubit]
        A mapping from index in the pauli string to `deltakit.circuit.Qubit` object.

    Returns
    -------
    list[_PauliGate]
        The paulistring as a list of PauliGate objects.
    """
    return [
        (PauliX if el == 1 else PauliZ if el == 3 else PauliY)(index_to_qubit[el_index])
        for el_index, el in enumerate(paulistr)
        if el > 0
    ]


def get_str_logical_operators_from_tableau(
    stabilisers: Collection[PauliString], num_logical_qubits: int | None = None
) -> list[tuple[PauliString, PauliString]]:
    """
    For a general stabiliser code, computes the logical operators for a collection of
    stabilisers.

    This method of computing the logical operators:
    - Guarantees the logical operators are independent,
    - Does NOT guarantee the logical operators are minimum-weight,
    - Does NOT guarantee for CSS codes that X logical operators are made purely of X gates
    and Z logical operators are made purely of Z gates (this is suspected but
    wasn't decidedly shown).

    From Stack Exchange post:
    https://quantumcomputing.stackexchange.com/questions/37812/how-to-find-a-set-of-independent-logical-operators-for-a-stabilizer-code-with-st

    Explanation from post:
    Solves for the observables as part of completing a tableau. It works by finding
    operations that turn the stabilisers into single-qubit terms. The observables are
    then created by looking at what undoing those operations turns the other qubits
    into.

    The stabilisers are provided as stim PauliStrings and so are the operators returned.

    Parameters
    ----------
    stabilisers : Collection[PauliString]
        The stabilisers as stim pauli string objects.
    num_logical_qubits : int, optional
        The number of logical qubits these stabilisers are expected to have. If provided,
        exactly this number of elements will be extracted from the end of the completed
        Tableau. If the number of logical qubits is known, providing it is safer
        because we are not 100% sure whether it is guaranteed that the stabilisers in
        the Tableau are exactly the same as the ones inputted (or whether they can be
        linear combinations of them, in which case the algorithm will think it is a
        logical operator). No example has been found where this was an issue so far, but
        this is to be extra safe. By default, None.

    Returns
    -------
    list[tuple[PauliString, PauliString]]
        The logical operators for the stabilisers provided. Each
        element in the list returned is an anticommuting pair
        of X and Z logical operators. Every other pair across and
        among commutes.
    """
    completed_tableau = Tableau.from_stabilizers(
        stabilisers,
        allow_redundant=True,
        allow_underconstrained=True,
    )

    iteration_range = range(len(completed_tableau))[::-1]

    if num_logical_qubits is not None:
        iteration_range = iteration_range[:num_logical_qubits]

    operators: list[tuple[PauliString, PauliString]] = []
    for k in iteration_range:
        z = completed_tableau.z_output(k)
        if z in stabilisers:
            break
        x = completed_tableau.x_output(k)
        operators.append((x, z))

    return operators


def get_logical_operators_from_tableau(
    stabilisers: Iterable[Stabiliser], num_logical_qubits: int | None = None
) -> tuple[tuple[set[_PauliGate], ...], tuple[set[_PauliGate], ...]]:
    """
    For a general stabiliser code, computes the logical operators for a collection of
    stabilisers.

    This method of computing the logical operators:
    - Guarantees the logical operators are independent,
    - Does NOT guarantee the logical operators are minimum-weight,
    - Does NOT guarantee for CSS codes that X logical operators are made purely of X gates
    and Z logical operators are made purely purely of Z gates (this is suspected but
    wasn't decidedly shown).

    The stabilisers are provided as Stabiliser objects and the operators returned are
    made of Pauli gates.

    Parameters
    ----------
    stabilisers : Iterable[Stabiliser]
        The stabilisers from which to generate logical operators.
    num_logical_qubits : int, optional
        The number of logical qubits these stabilisers are expected to have. If provided,
        exactly this number of elements will be extracted from the end of the completed
        Tableau. If the number of logical qubits is known, providing it is safer
        because we are not 100% sure whether it is guaranteed that the stabilisers in
        the Tableau are exactly the same as the ones inputted (or whether they can be
        linear combinations of them, in which case the algorithm will think it is a
        logical operator). No example has been found where this was an issue so far, but
        this is to be extra safe. By default, None.

    Returns
    -------
    tuple[tuple[set[_PauliGate], ...], tuple[set[_PauliGate], ...]]
        The logical operators, provided as a tuple of all the X logical
        operators at index 0 and all the Z logical operators at index 1.
        The logical operators are ordered in anticommuting pairs, such
        that the ith X logical commutes with all X and Z logical operators,
        except for the ith Z logical operator, with which it anticommutes.
    """
    # compute the mapping from qubit to index in the pauli string
    qubit_to_pauli_index: dict[Qubit, int] = {}
    index = 0
    for stabiliser in stabilisers:
        for pauli in stabiliser.paulis:
            if (pauli is not None) and (
                (qubit := pauli.qubit) not in qubit_to_pauli_index
            ):
                qubit_to_pauli_index[qubit] = index
                index += 1

    # convert the stabilisers to paulistring format
    paulistrings = [
        pauli_gates_to_stim_pauli_string(stabiliser.paulis, qubit_to_pauli_index)
        for stabiliser in stabilisers
    ]

    pauli_index_to_qubit = {v: k for k, v in qubit_to_pauli_index.items()}

    # compute the logical operators as paulistrings
    str_operators = get_str_logical_operators_from_tableau(
        paulistrings, num_logical_qubits
    )

    x_str_operators, z_str_operators = (
        list(zip(*str_operators)) if len(str_operators) > 0 else ([], [])
    )

    # convert paulistrings to operator format
    x_operators = tuple(
        set(paulistring_to_operator(opr, pauli_index_to_qubit))
        for opr in x_str_operators
    )
    z_operators = tuple(
        set(paulistring_to_operator(opr, pauli_index_to_qubit))
        for opr in z_str_operators
    )

    return x_operators, z_operators


def get_logical_operators_from_css_parity_check_matrices(
    hx: NDArray, hz: NDArray, column_to_qubit: dict[int, Qubit]
) -> tuple[tuple[set[PauliX], ...], tuple[set[PauliZ], ...]]:
    """
    For a CSS stabiliser code, computes the logical operators using its parity check
    matrices and the BPOSD package (https://arxiv.org/abs/2005.07016).

    This method of computing the logical operators:
    - Does NOT guarantee the logical operators are independent,
    - Does NOT guarantee the logical operators are minimum-weight,
    - Guarantees for CSS codes that X logical operators are made purely of X gates
    and Z logical operators are made purely purely of Z gates (this is suspected but
    wasn't decidedly shown).

    Parameters
    ----------
    h_x : NDArray
        The check matrix (containing only 0 and 1) for X stabilisers where each
        row represents an X stabiliser. If an empty matrix, then this means the
        CSS code has no X stabilisers.
    h_z : NDArray
        The check matrix (containing only 0 and 1) for Z stabilisers where each
        row represents a Z stabiliser. If an empty matrix, then this means the
        CSS code has no Z stabilisers.
    column_to_qubit : dict[int, Qubit]
        A mapping from column index in the parity check matrices h_x, h_z to the
        data qubit the column's entries describe.

    Returns
    -------
    tuple[tuple[set[PauliX], ...], tuple[set[PauliZ], ...]]
        The logical operators, provided as a tuple of all the X logical operators
        at index 0 and all the Z logical operators at index 1.
    """
    x_logs, z_logs = css_code_compute_logicals(hx, hz)

    return tuple(
        {PauliX(column_to_qubit[i]) for i, x in enumerate(log_op) if x}
        for log_op in x_logs
    ), tuple(
        {PauliZ(column_to_qubit[i]) for i, x in enumerate(log_op) if x}
        for log_op in z_logs
    )


def css_code_compute_logicals(
    hx: NDArray[np.floating], hz: NDArray[np.floating]
) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
    """Drop-in replacement for calling bposd.css_code.compute_logicals.

    Note:
        This function has been taken from the following repository under the MIT
        licence: [bp_osd repository](https://github.com/quantumgizmos/bp_osd).

        The following modifications were performed:
            1. Add more detailed typing information (parameter types, return type).
            2. Remove two lines of code that were not directly used by the method:
                ```
                if self.K == np.nan:
                    self.compute_dimension()
                ```
            3. Change the names of the parameters of compute_lz to avoid a name clash.
            4. Remove all the `self.`.
            5. Add a docstring.
            6. Add typing to the internal `compute_lz` function.
            7. Only use dense matrices because the inputs are dense anyway.

        You can check the original version of this function at
        [this permalink](https://github.com/quantumgizmos/bp_osd/blob/8894ec654b24ae875c07e5a361dcae9a77d748ce/src/bposd/css.py#L75).

    Args:
        hx: parity check matrix for the X code.
        hz: parity check matrix for the Z code.

    Returns:
        a tuple ``(lx, lz)`` representing the X and Z logicals.
    """
    def compute_lz(_hx: NDArray[np.floating], _hz: NDArray[np.floating]) -> NDArray[np.floating]:
        # lz logical operators
        # lz\in ker{hx} AND \notin Im(Hz.T)

        # compute the kernel basis of hx
        # Note that because inputs are dense arrays, it is fine for every array to be dense in this
        # function.
        ker_hx = mod2.nullspace(_hx).todense()
        # Row reduce to find vectors in kx that are not in the image of hz.T.
        log_stack = np.vstack([_hz, ker_hx])

        rank_hz = mod2.rank(_hz)
        pivots = mod2.pivot_rows(log_stack)[rank_hz:]

        return np.asarray(log_stack[pivots])

    return compute_lz(hz, hx), compute_lz(hx, hz)



# (c) Copyright Riverlane 2020-2025.
"""Description of ``deltakit.explorer.enums`` namespace here."""

from deltakit_explorer.enums._api_enums import (
                                                APIEndpoints,
                                                DataFormat,
                                                DecoderType,
                                                QECECodeType,
                                                QECExperimentType,
)

# List only public members in `__all__`.
__all__ = [s for s in dir() if not s.startswith("_")]



from __future__ import annotations

from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt
from deltakit_circuit._circuit import Circuit

from deltakit_explorer.analysis.error_budget._gradient import inverse_lambda_gradient_at
from deltakit_explorer.analysis.error_budget._memory import (
    MemoryGenerator,
    get_rotated_surface_code_memory_circuit,
)
from deltakit_explorer.analysis.error_budget._parameters import (
    FittingParameters,
    SamplingParameters,
)


@dataclass
class ErrorBudgetResult:
    """Result of an error budgeting computation.

    Attributes:
        contributions: contributions for each of the noise parameters to the error budget.
        contribution_stddevs: estimation of the standard deviation of each of the ``contributions``.
    """

    contributions: tuple[float, ...]
    contribution_stddevs: tuple[float, ...]

    @property
    def lambda_estimate(self) -> float:
        """Returns the estimation of Λ according to the computed budget."""
        return float(np.sum(self.contributions))

    @property
    def lambda_stddev_estimate(self) -> float:
        """Returns an estimation of the standard deviation on Λ according to the computed budget."""
        return float(np.sqrt(np.sum(np.asarray(self.contribution_stddevs) ** 2)))

    @staticmethod
    def from_gradient(
        gradient: npt.NDArray[np.floating],
        gradient_stddevs: npt.NDArray[np.floating],
        noise_parameters: npt.NDArray[np.floating],
    ) -> ErrorBudgetResult:
        """Create an instance from gradient and noise parameters."""
        contributions = np.abs(gradient * noise_parameters)
        stddevs = np.abs(gradient_stddevs * noise_parameters)
        return ErrorBudgetResult(
            tuple(map(float, contributions.ravel())), tuple(map(float, stddevs.ravel()))
        )


def get_error_budget(
    noise_model: Callable[[Circuit, npt.NDArray[np.floating]], Circuit],
    noise_parameters: npt.NDArray[np.floating] | Sequence[float],
    num_rounds_by_distances: Mapping[int, Sequence[int]],
    noise_parameters_exploration_bounds: list[tuple[float, float]],
    fitting_parameters: FittingParameters = FittingParameters(),
    sampling_parameters: SamplingParameters = SamplingParameters(),
    memory_generator: MemoryGenerator
    | Mapping[int, Mapping[int, Circuit]] = get_rotated_surface_code_memory_circuit,
) -> ErrorBudgetResult:
    """Compute the error budget of the provided ``noise_model``.

    Args:
        noise_model (Callable[[Circuit, npt.NDArray[np.floating]], Circuit]): a callable
            adding noise to the provided circuit, according to the parameters provided.
        noise_parameters (npt.NDArray[numpy.floating] | Sequence[float]): valid
            parameters to forward to ``noise_model`` representing the point at which the
            gradient should be computed.
        num_rounds_by_distances (Mapping[int, Sequence[int]]): a mapping from each code
            distance that should be tested to the number of rounds that should be
            sampled in order to estimate the logical error-probability per round, to
            ultimately get 1 / Λ.
        noise_parameters_exploration_bounds (list[tuple[float, float]]): ``(min, max)``
            bounds for each noise parameter of the provided ``noise_model``. A degree
            ``fitting_degree`` polynomial will be fitted on the interval ``[min, max]``.
            The corresponding noise parameter from the provided ``noise_model`` should
            be strictly contained in ``[min, max]`` (i.e., for any valid ``i``, the
            following is true:
            ``noise_parameters_exploration_bounds[i][0] <
            noise_model.noise_parameters[i] <
            noise_parameters_exploration_bounds[i][1]``). Ideally, the lower (resp.
            upper) bound provided must be such that the logical error probability when
            replacing the parameter with its lower (resp. upper) bound is above
            ``100 / max_shots`` to ensure enough fails are observed with ``max_shots``
            shots (resp. below ``1 / 2`` to ensure that we can compute the logical error
            probability per round).
        fitting_parameters: additional parameters relating to how the gradient is
            estimated.
        sampling_parameters: additional parameters relating to the sampling tasks used to
            estimate 1 / Λ indirectly.
        memory_generator (MemoryGenerator): a callable that can generate a memory
            experiment. The resulting circuit will go through the provided
            ``noise_model`` for different values of the noise parameters.

    Returns:
        the error-budgeting result, which consists of an array of contributions for each
        of the noise parameters of the provided ``noise_model`` along with their
        associated standard deviations.
    """
    # We will compute the gradient at the half point following the methodology outlined in
    # https://doi.org/10.1038/s41586-021-03588-y (Supplementary materials, Section VIII.C.).
    point = np.asarray(noise_parameters) / 2
    # Evaluate the gradient.
    gradient, gradient_stddev = inverse_lambda_gradient_at(
        noise_model,
        point,
        num_rounds_by_distances,
        noise_parameters_exploration_bounds,
        fitting_parameters,
        sampling_parameters,
        memory_generator,
    )
    return ErrorBudgetResult.from_gradient(
        gradient, gradient_stddev, np.asarray(noise_parameters)
    )
